/*
 ** xkaapi
 ** 
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 ** thierry.gautier@inrialpes.fr
 ** fabien.lementec@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#ifndef KAAPIC_H_INCLUDED
# define KAAPIC_H_INCLUDED

#include <stddef.h>
#include <stdint.h>
#include "kaapi.h"
#include "kaapi_cpuset.h"

#if defined(__cplusplus)
extern "C" {
#endif
  
/*
 kaapic_spawn(attr, int args, body, [type, []])
 kaapic_task_register(ints nargs, void* body_cpu, (type[])*)
 kaapic_task_impl_add(body_cpu, type, body_gpu)
 kaapic_task_spawn(attr, body_cpu, ()*)
 */

/* This is the portable C interface the kaapi runtime */
enum kaapic_type
{
  KAAPIC_TYPE_CHAR,
  KAAPIC_TYPE_SCHAR,
  KAAPIC_TYPE_SHRT,
  KAAPIC_TYPE_INT,
  KAAPIC_TYPE_LONG,
  KAAPIC_TYPE_LLONG,
  KAAPIC_TYPE_INT8,
  KAAPIC_TYPE_INT16,
  KAAPIC_TYPE_INT32,
  KAAPIC_TYPE_INT64,
  KAAPIC_TYPE_UCHAR,
  KAAPIC_TYPE_USHRT,
  KAAPIC_TYPE_UINT,
  KAAPIC_TYPE_ULONG,
  KAAPIC_TYPE_ULLONG,
  KAAPIC_TYPE_UINT8,
  KAAPIC_TYPE_UINT16,
  KAAPIC_TYPE_UINT32,
  KAAPIC_TYPE_UINT64,
  KAAPIC_TYPE_FLT,
  KAAPIC_TYPE_DBL,
  KAAPIC_TYPE_LDBL,
  KAAPIC_TYPE_PTR,
  _KAAPIC_TYPE_MAX
};

enum kaapic_mode
{
  KAAPIC_MODE_R = 0,
  KAAPIC_MODE_W = 1,
  KAAPIC_MODE_RW= 2,
  KAAPIC_MODE_X = KAAPIC_MODE_RW,
  KAAPIC_MODE_CW= 3,
  KAAPIC_MODE_V = 4,
  KAAPIC_MODE_T = 5,  /* temporary */
  KAAPIC_MODE_S = 6,  /* stack */
  _KAAPIC_MODE_MAX = 7   /* last mode */
};


enum kaapic_reducop_t
{
  KAAPIC_REDOP_VOID = 0,
  KAAPIC_REDOP_PLUS = 1,
  KAAPIC_REDOP_MUL  = 2,
  KAAPIC_REDOP_MINUS= 3,
  KAAPIC_REDOP_AND  = 4, /* & */
  KAAPIC_REDOP_OR   = 5, /* | */
  KAAPIC_REDOP_XOR  = 6, /* ^ */
  KAAPIC_REDOP_LAND = 7, /* && */
  KAAPIC_REDOP_LOR  = 8, /* || */
  KAAPIC_REDOP_MAX  = 9, 
  KAAPIC_REDOP_MIN  = 10 
};

enum kaapic_init_flag {
  KAAPIC_START_ALL = 0,
  KAAPIC_START_ONLY_MAIN = 1
};

/** kaapic_init INIT initializes the runtime. 
  It must be called once per pro- gram before using any of 
  the other routines. 
  If successful, there must be a corresponding 
  kaapic_finalize at the end of the program.

  \param flags: if not zero, start only the main thread 
  to avoid disturbing the execution until tasks are actually scheduled. 
  The other threads are suspended waiting for a parallel 
  region to be entered (refer to KAAPIF BEGIN PARALLEL)
*/
extern int kaapic_init(int32_t flags);

/** Finalizes the Kaapi runtime.
*/
extern int kaapic_finalize(void);

/** Returns the current time in micro-second.
*/
extern double kaapic_get_time(void);

/** Returns the number of parallel thread available to the X-Kaapi runtime
*/
extern int32_t kaapic_get_concurrency(void);

/* Returns the current thread identifier. 
   Note it should only be called in the context of a X-Kaapi thread, else
   -1 is returned.
*/
extern int32_t kaapic_get_thread_num(void);


/* Scheduling strategy for foreach loop
*/
typedef enum {
  KAAPIC_FOREACH_SCHED_STEAL      = 0x00,  /* steal half strategy */
  KAAPIC_FOREACH_SCHED_DISTRIBUTE = 0x01,   /* add to the default strategy a distribution step */
  KAAPIC_FOREACH_SCHED_DEFAULT    = KAAPIC_FOREACH_SCHED_DISTRIBUTE
} kaapic_foreach_attr_policy_t;

/* Internal type, not documented to the user
*/
typedef enum {
  KAAPIC_DATADIST_VOID       = 0,
  KAAPIC_DATADIST_BLOCCYCLIC = 1
} _kaapic_foreach_attr_datadist_type_t;

/* Internal type, not documented to the user
*/
typedef struct _kaapic_foreach_attr_datadist_t {
  _kaapic_foreach_attr_datadist_type_t type;      /* discriminant for the distribution */
  union {
    struct { /* information about bloc-cyclic distribution of data */
      unsigned long long size;
      unsigned int  length; 
    } bloccyclic;
  } dist;
} _kaapic_foreach_attr_datadist_t;

/* Foreach attribut.
   Default values are :
   * sgrain, pgrain = 1
   * policy         = KAAPIC_FOREACH_SCHED_DEFAULT
   * nthreads       = depending of the runtime number of threads, no limit
   * threadset      = no constraints
*/
typedef struct kaapic_foreach_attr_t {
  unsigned int         init;      /* flags for each attributes == 1 iff initialize */
  union {
    struct {
      long                 s_grain;  /* default = 1 */
      long                 p_grain;  /* default = 1 */
    } li;
    struct { /* ull case */
      unsigned long long   s_grain;  /* default = 1 */
      unsigned long long   p_grain;  /* default = 1 */
    } ull;
  }                            rep;
  kaapic_foreach_attr_policy_t policy;    /* choose the policy for splitting */
  unsigned int                 nthreads;  /* number of threads for initial splitting */
  kaapi_cpuset_t               threadset; /* thread set used for initial distribution i = kid */
} kaapic_foreach_attr_t;
  
/*
*/
extern int kaapic_foreach_attr_init(kaapic_foreach_attr_t* attr);

/*
*/
extern int kaapic_foreach_attr_set_policy(
  kaapic_foreach_attr_t* attr, 
  kaapic_foreach_attr_policy_t policy
);

/*
*/
extern int kaapic_foreach_attr_get_policy(
  const kaapic_foreach_attr_t* attr,
  kaapic_foreach_attr_policy_t* policy
);

/*
*/
extern int kaapic_foreach_attr_set_grains(
  kaapic_foreach_attr_t* attr, 
  long s_grain,
  long p_grain
);

/*
*/
extern int kaapic_foreach_attr_get_grains(
  const kaapic_foreach_attr_t* attr,
  long* s_grain,
  long* p_grain
);

/*
*/
extern int kaapic_foreach_attr_set_grains_ull(
  kaapic_foreach_attr_t* attr, 
  unsigned long long s_grain,
  unsigned long long p_grain
);

/*
*/
extern int kaapic_foreach_attr_get_grains_ull(
  const kaapic_foreach_attr_t* attr,
  unsigned long long* s_grain,
  unsigned long long* p_grain
);

/* Set/Get Nthreads attribute
*/
extern int kaapic_foreach_attr_set_nthreads(
  kaapic_foreach_attr_t* attr, 
  unsigned int nthreads
);

extern int kaapic_foreach_attr_get_nthreads(
  const kaapic_foreach_attr_t* attr,
  unsigned int* nthreads
);

/* Set/Get cpuset attribute
*/
extern int kaapic_foreach_attr_set_cpuset(
  kaapic_foreach_attr_t* attr, 
  const kaapi_cpuset_t* threadset
);

extern int kaapic_foreach_attr_get_cpuset(
  kaapic_foreach_attr_t* attr, 
  kaapi_cpuset_t*        threadset
);


/* Specify the distribution of initial range [0,N) over the NUMA node
*/
extern int kaapic_foreach_attr_set_bloccyclic_datadistribution(
  kaapic_foreach_attr_t* attr, 
  unsigned long long blocsize,
  unsigned int cyclelength
);

/*
*/
static inline int kaapic_foreach_attr_destroy(kaapic_foreach_attr_t* attr)
{ return 0; }

/* See documentation
*/
extern void kaapic_foreach( 
  int32_t first, 
  int32_t last, 
  const kaapic_foreach_attr_t* attr,
  int32_t nparam, 
  /* void (*body)(int32_t, int32_t, int32_t, ...), */
  ...
);


/* See documentation
*/
extern void kaapic_foreach_ull( 
  unsigned long long first, 
  unsigned long long last, 
  const kaapic_foreach_attr_t* attr,
  int32_t nparam, 
  /* void (*body)(unsigned long long, unsigned long long, int32_t, ...), */
  ...
);


/* See documentation
*/
extern void kaapic_foreach_with_format(
  int32_t first, 
  int32_t last, 
  const kaapic_foreach_attr_t* attr,
  int32_t nparam, 
  /* void (*body)(int32_t, int32_t, int32_t, ...), */
  ...
);

extern void kaapic_foreach_with_format_ull(
  unsigned long long first, 
  unsigned long long last, 
  const kaapic_foreach_attr_t* attr,
  int32_t nparam, 
  /* void (*body)(unsigned long long, unsigned long long, int32_t, ...), */
  ...
);

/* Allocate a data with the same scope as a task, ie. in the Kaapi'thread stack
*/
extern void* kaapic_alloca( size_t sz );


typedef enum kaapic_task_flag_t {
  KAAPIC_TASK_S_DEFAULT,
  KAAPIC_TASK_UNSTEALABLE,
  KAAPIC_TASK_LOCDOM
} kaapic_task_flag_t;

#define   KAAPIC_ARCH_DEFAULT     ((1U << KAAPI_PROC_TYPE_HOST) | (1U << KAAPI_PROC_TYPE_CUDA))
#define   KAAPIC_ARCH_CPU_ONLY    (1U << KAAPI_PROC_TYPE_HOST)
#define   KAAPIC_ARCH_GPU_ONLY    (1U << KAAPI_PROC_TYPE_CUDA)  
  
/*
*/
typedef struct kaapic_spawn_attr_t {
  int										kid;     /* -1 if any kprocessor */
  kaapic_task_flag_t   flag;
  uint8_t              prio;
  uint8_t              arch;    /* task architecture */
  void                (*alter_body)(); /* alternative body (GPU) */
} kaapic_spawn_attr_t;
  
/*
*/
extern int kaapic_spawn_attr_init(kaapic_spawn_attr_t* attr);

/*
*/
static inline int kaapic_spawn_attr_destroy(kaapic_spawn_attr_t* attr)
{ return 0; }

/* Specify that task should be push into the queue of the kid kprocessor.
   If kid is bigger than the number of active kprocessor then the kid is
   considered as modulo the number of active kprocessor.
*/
extern int kaapic_spawn_attr_set_ld(
  kaapic_spawn_attr_t* attr, 
  int localitydomain
);

/* Set alternative body for GPUs */
extern int kaapic_spawn_attr_set_alternative_body(
                                       kaapic_spawn_attr_t* attr,
                                       void (*body)()
                                       );
  
extern int kaapic_spawn_attr_set_arch(
                               kaapic_spawn_attr_t* attr,
                               uint8_t arch
                               );

/** Set task priority.
 \param attr: task attribute.
 \param prio: priority value.
 \retval 0 in case of success, else an error code.
*/
extern int kaapic_spawn_attr_set_prio( kaapic_spawn_attr_t* attr, uint8_t prio );
  
/* Specify flag on task:
   - UNSTEALABLE: task cannot be steal
   - LOCDOM: task must be executed on the specified locality domaine (kaapic_spawn_attr_set_ld)
*/
extern int kaapic_spawn_attr_flags(
  kaapic_spawn_attr_t* attr,
  kaapic_task_flag_t flag
);


/* Create a task that may steal.
   See documentation for restriction on by value passing rule.
   The attribut attr may be null and correspond to the default attribut.
   \param body : the task body. 
   \param nargs: the argument count. 
   \param ...: a { MODE, TYPE, [REDOP, ]COUNT, VALUE/@ } tuple list.
   \retval 0 in case of success, else an error code
*/
extern int kaapic_spawn(const kaapic_spawn_attr_t* attr, int32_t nargs,
			void (*body)(), ...);

/*
*/
extern void kaapic_sync(void);
  
extern void kaapic_memory_sync(void);


/* Flag for parallel regions
*/
enum {
  KAAPIC_FLAG_STATIC_SCHED = 0x1,
  KAAPIC_FLAG_END_NOSYNC   = 0x2,  /* no implicit sync */
  KAAPIC_FLAG_DEFAULT      = 0x0
};

/* Start parallel region with flag.
   \retval 0 in case of success, else an error code
*/
extern int kaapic_begin_parallel(int flags);

/* End parallel region with flag.
   If last parallel_begin has STATIC_SCHED flag, then
   the STATIC_SCHED_FLAG must also be passed to end_parallel.
   \retval 0 in case of success, else an error code
*/
extern int kaapic_end_parallel(int flags);

#if defined(__cplusplus)
}
#endif

#endif /* KAAPIC_H_INCLUDED */
