/*
 ** xkaapi
 ** 
 ** Created on Tue Mar 31 15:19:14 2009
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 ** thierry.gautier@inrialpes.fr
 ** fabien.lementec@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
 
/* Note on the implementation.
   The work structure is decomposed in two parts that follows the two
   scheduling level adopted here.
   At the upper level, the work is a kaapic_global_work_t. It represents
   the iteration space distributed among a set of participants.
   At the lower lelve, the work is a kaapic_local_work_t which is
   represented as a kaapi_workqueue_t.
   Initialisation of the global work must be made by a unique caller thread.
   Once created, the caller thread has initialized the global work and the
   local work information. The adaptive task publishes the splitter that will
   create new participant.
   
   Once a participant is enrolled to work on the global work instance, i.e. after
   he begins to execute the _kaapic_thief_entrypoint, he never returns until the
   global end of the work.
   This optimization allows to bypass a return to the scheduler. Nevertheless, 
   it can let work inactive until all the enrolled threads complete.
   
   The distribution is not an foreach_attribute, but it must be.
   The current distribution data structure is handle by the wa field of the global
   work data structure. A thread with given tid, correspond to initial slice 
   [startindex[pos]..starindex[pos+1]) where pos = tid2pos[tid].
   The table tid2pos is used to compact information at the begin of the array
   startindex. Only thread with tid such that the tid-th bit is set in the
   map field can steal the initial slice.
   T.G.
*/
 
#include "kaapi_impl.h"
#include "kaapic_impl.h"

#define _GNU_SOURCE
#include <sched.h>


#define FATAL()						\
do {							\
  printf("fatal error @ %s::%d\n", __FILE__, __LINE__);	\
  exit(-1);						\
} while(0)


#if defined(KAAPI_DEBUG)
# define PRINTF(__s, ...) printf(__s, __VA_ARGS__)
#else
# define PRINTF(__s, ...)
#endif

#define KAAPI_LOCAL_LOG(x)
//#define KAAPI_LOCAL_LOG(x) x


static int kaapic_foreach_globalwork_next(
  kaapic_local_work_t*     lwork,
  kaapi_workqueue_index_t* first,
  kaapi_workqueue_index_t* last
);


/* Call to fill localwork of kprocessors id tid:
   The function uses wa.map and wa.tid2pos bitmap to try first to pop the reserved slice.
   If wa.map[tid] is not set then it tries to pop the next reserved slices in the bitmap.
   
   If flag ==1 steal any slice is not reserved
*/
int kaapic_global_work_pop
(
  kaapic_global_work_t* gw,
  kaapi_processor_id_t tid, 
  kaapi_workqueue_index_t* i, 
  kaapi_workqueue_index_t* j,
  int flag 
)
{
  int retval;
  int idx;
  int pos;

  kaapi_assert_debug(tid<KAAPI_MAX_PROCESSOR);

  /* first tries to pop position at tid */
  pos = gw->wa.tid2pos[tid];
  if ((pos == (uint8_t)-1) && (flag ==1))
  {
repop_any:
    /* no reserved slice: steal one */
    idx = kaapi_bitmap_first1_and_zero(&gw->wa.map);
    if (idx !=0)
    {
      tid = idx-1;
      pos = gw->wa.tid2pos[tid];
      *i  = gw->wa.startindex[pos];
      *j  = gw->wa.startindex[pos+1];
      kaapi_assert_debug( *i <*j );
KAAPI_LOCAL_LOG( printf("[%i]:: [kaapic_global_work_pop] pop first bloc tid:%i, [%i, %i)\n", kaapi_get_self_kid(), tid, *i, *j); fflush(stdout); )
      return 1;
    }
    *i = *j = 0;
    return 0;
  }
  kaapi_assert_debug( pos >= 0 );
  kaapi_assert_debug( pos < kaapi_getconcurrency() );
  
  retval = (0 == kaapi_bitmap_unset(&gw->wa.map, tid));
  if (!retval)
  {
    pos = -1;
    goto repop_any;
  }
  pos = gw->wa.tid2pos[tid];
  *i = gw->wa.startindex[pos];
  *j = gw->wa.startindex[pos+1];
  KAAPI_LOCAL_LOG( printf("[%i]:: [kaapic_global_work_pop] pop bloc tid:%i, [%i, %i)\n", kaapi_get_self_kid(), tid, *i, *j ); fflush(stdout); )

  kaapi_assert_debug( *i <*j );
  return 1;
}


/* return !=0 iff successful steal op */
static int kaapic_global_work_steal
(
  kaapic_global_work_t*    gwork,
  kaapi_processor_t*       kproc, 
  kaapi_workqueue_index_t* i, 
  kaapi_workqueue_index_t* j
)
{
  kaapi_workqueue_index_t unit_size;
  kaapi_workqueue_index_t range_size;
  int retval;

#if defined(KAAPI_USE_PERFCOUNTER)
  ++KAAPI_PERF_REG(kproc, KAAPI_PERF_ID_STEALREQ);
#endif

  /* here to do / to test:
     - any try to pop a slice closed to the tid of the thread
     - only 0 can pop a non poped slice
  */
#if defined(KAAPIC_ALLOWS_WORKER_STEAL_SLICE) //  Empeche le vol d'un slice 
  /* caller has already pop and finish its slice, if it is 0 then may pop
     the next non null entry
  */
#if 0
  int tid = kproc->kid;
  if (tid == 0)
  {
    kaapi_assert_debug(tid<KAAPI_MAX_PROCESSOR);
#else
  if (1)
  {
#endif
    int tidpos = kaapi_bitmap_first1( &gwork->wa.map );
    if ((tidpos !=0) && kaapic_global_work_pop(gwork, tidpos-1, i, j, 1 ))
    {
#if defined(KAAPI_USE_PERFCOUNTER)
      ++KAAPI_PERF_REG(kproc, KAAPI_PERF_ID_STEALREQOK);
#endif
      kaapi_assert_debug( *i < *j );
      KAAPI_LOCAL_LOG( printf("[%i]:: [kaapic_global_work_steal] steal [%i, %i)\n", kaapi_get_self_kid(), *i, *j); fflush(stdout); )

      /* success */
      return 1;
    }
  }
#endif
    
  /* try to steal from a random selected local workqueue */
  /* the code here is closed to the emitsteal code except we bypass the request mechanism */
  kaapi_victim_t  victim;
  int             err;

redo_select:
  /* warning here the method does not return except: work is found or total work is done
  */
  if (KAAPI_ATOMIC_READ(&gwork->workremain) ==0) 
    return 0;
  kaapi_slowdown_cpu();

  /* select the victim processor */
  err = (*kproc->fnc_select)( kproc, &victim, KAAPI_SELECT_VICTIM );
  if (unlikely(err !=0)) 
    goto redo_select;

  /* never pass by this function for a processor to steal itself */
  if (kproc == victim.kproc) 
  {
    (*kproc->fnc_select)( kproc, &victim, KAAPI_STEAL_FAILED );
    goto redo_select;
  }
    
  /* do not steal if range size <= par_grain */
  kaapic_local_work_t* lwork = &gwork->lwork[victim.kproc->kid];

  /* because object lwork may exist without beging initialized, return if not */
  if (lwork->init == 0)
  {
    (*kproc->fnc_select)( kproc, &victim, KAAPI_STEAL_FAILED );
    goto redo_select;
  }

  /* try to steal the local work */
  const kaapic_work_info_t* const wi = &gwork->wi;

  range_size = kaapi_workqueue_size(&lwork->cr);
  if (range_size <= wi->rep.li.par_grain)
  {
    (*kproc->fnc_select)( kproc, &victim, KAAPI_STEAL_FAILED );
    goto redo_select;
  }

  unit_size = range_size / 2;
  if (unit_size ==0) unit_size = 1;

  _kaapi_workqueue_lock(&lwork->cr);
  retval = kaapi_workqueue_steal(&lwork->cr, i, j, unit_size);
  _kaapi_workqueue_unlock(&lwork->cr);
  kaapi_assert_debug( (retval!=0) || (*i <= *j) );

  if (retval ==0)
  {
#if defined(KAAPI_USE_PERFCOUNTER)
    ++KAAPI_PERF_REG(kproc, KAAPI_PERF_ID_STEALREQOK);
#endif
    kaapi_assert_debug( (*i < *j) && ((*j-*i) <= unit_size) );
    KAAPI_LOCAL_LOG( printf("[%i]:: [kaapic_global_work_steal] steal %i: [%i, %i)\n", kaapi_get_self_kid(), lwork->tid, *i, *j); fflush(stdout); )
    (*kproc->fnc_select)( kproc, &victim, KAAPI_STEAL_SUCCESS );
    return 1;
  }

  (*kproc->fnc_select)( kproc, &victim, KAAPI_STEAL_FAILED );

  goto redo_select;
}


/* fwd decl */
static void _kaapic_thief_entrypoint(void*, kaapi_thread_t*,  kaapi_task_t* );


/* Used to start parallel region if required */
extern unsigned int kaapic_do_parallel;


/* Initial work distribution wa.
   - self_tid is the thread that initial the globalwork, it will get range between
   startindex[self_tid]=0 and startindex[self_tid+1]=chuncksize
   - [first,last) the slice to distribute
*/
static void _kaapic_foreach_initwa(
  kaapic_work_distribution_t* wa,
  int                         policy,
  int                         self_tid,
  int                         nthreads,
  const kaapi_bitmap_value_t* threadset,
  kaapi_workqueue_index_t     first,
  kaapi_workqueue_index_t     last
)
{
  kaapi_workqueue_index_t range_size;
  long off;
  long scale;
  int finalsize;
  int i;

  int concurrency;
  kaapi_assert_debug(KAAPI_MAX_PROCESSOR < (uint16_t)~0 );
  
  kaapi_bitmap_clear( &wa->map );
  concurrency = kaapi_getconcurrency();

  if (policy == KAAPIC_FOREACH_SCHED_STEAL)
  {
    kaapi_bitmap_set( &wa->map, self_tid);
    for (i=0; i<concurrency; ++i)
      wa->tid2pos[i] = (uint16_t)-1; /* not in the set */
    wa->tid2pos[self_tid] = 0;
    wa->startindex[0] = first;
    wa->startindex[1] = last;
    return;
  }
  
  /* compute the map is used iff cpuset attribute is used*/

  /* concurrency is defined as the size of the set */
  finalsize = nthreads;

  /* split the range in equal slices */
  range_size = last - first;

  /* handle concurrency too high case */
  if (range_size < nthreads)
    finalsize = (int)range_size;

  /* round range to be multiple of nthreads 
     self_tid will get the biggest part (the first one)
     Distribution is a block distribution. Any other distribution should be used.
  */
  off = range_size % finalsize;
  range_size -= off;
  scale = range_size / finalsize;
  
  /* init logical mapping from tid to [0, ..., n] such that localtid is attached to 0 if
     is in the set.
     Initialize also all the localwork to empty (lock is initialized !)
  */
  uint16_t localcount = 0;
  wa->tid2pos[self_tid] = localcount++;
  kaapi_bitmap_set( &wa->map, self_tid);

  for (i=0; i<concurrency; ++i)
  {
    if (i == self_tid)
      continue;
    if (kaapi_bitmap_value_get(threadset, i)  && (localcount < finalsize))
    {
      wa->tid2pos[i] = localcount++;
      kaapi_bitmap_set( &wa->map, i);
    }
    else 
      wa->tid2pos[i] = (uint16_t)-1; /* not in the set */
  }

  /* fill the start indexes: here it should be important to allocate slices depending 
     of the futur thread id... ?
  */
  wa->startindex[0] = first;
  wa->startindex[1] = first+off+scale;
  for (i=1; i<finalsize; ++i)
    wa->startindex[i+1] = wa->startindex[i]+scale;
  for (i=finalsize; i<concurrency; ++i)
    wa->startindex[i+1] = wa->startindex[finalsize];

#if 0
  for (i=0; i<concurrency; ++i)
  {
    int pos= wa->tid2pos[i];
    if (pos != -1)
      printf("Thread %i, cpuid: %i,  initial work: [%d, %d[\n", i, kaapi_all_kprocessors[i]->cpuid, wa->startindex[pos], wa->startindex[pos+1]);
    else
      printf("Thread %i empty initial work\n", i);
  }
#endif

  kaapi_assert_debug(wa->startindex[finalsize] == last);
}


/* Initialize the global work:
   - push a frame of the calling thread
   - allocate the global work into the stack
*/
kaapic_global_work_t* kaapic_foreach_global_workinit
(
  kaapi_thread_context_t* self_thread,
  kaapi_workqueue_index_t first, 
  kaapi_workqueue_index_t last,
  const kaapic_foreach_attr_t*  attr,
  kaapic_foreach_body_t   body_f,
  kaapic_body_arg_t*      body_args
)
{
  kaapic_global_work_t* gwork;
  kaapic_local_work_t* lwork;
  int concurrency;
  int i;

  /* if empty gwork 
  */
  if (first >= last) 
    return 0;
  
  /* push a new frame.
     will be poped at the terminaison 
  */
  kaapi_thread_push_frame_( self_thread );
  int localtid = self_thread->stack.proc->kid;

  kaapi_thread_t* const thread = kaapi_threadcontext2thread(self_thread);
  
  /* allocate the gwork */
  gwork = kaapi_thread_pushdata_align(thread, sizeof(kaapic_global_work_t), KAAPI_MAX_DATA_ALIGNMENT);
  kaapi_assert_debug(gwork !=0);

  concurrency = kaapi_getconcurrency();

  /* Get attribut's values */
  if (attr ==0)
  { /* default attributs, report them in work info */
    gwork->wi.flags         = 0;
    gwork->wi.policy        = KAAPIC_FOREACH_SCHED_DEFAULT;
    gwork->wi.rep.li.par_grain = 1;
    gwork->wi.rep.li.seq_grain = 1;
    gwork->wi.nthreads      = concurrency;
    /* accept requests from any kids */
    kaapi_bitmap_value_set_low_bits(&gwork->wi.threadset, concurrency);
  }
  else
  {
    gwork->wi.flags         = attr->init;
    kaapic_foreach_attr_get_policy( attr, &gwork->wi.policy );
    kaapic_foreach_attr_get_grains( attr, &gwork->wi.rep.li.seq_grain, &gwork->wi.rep.li.par_grain );
    kaapic_foreach_attr_get_nthreads( attr, &gwork->wi.nthreads );
    if (gwork->wi.nthreads > concurrency)
      gwork->wi.nthreads = concurrency;

    /* because sizeof(kaapi_cpuset_t) may differ from sizeof(kaapi_bitmap_value_t),
       we need to manage by hand, the copy.
       One assertion that is always true: sizeof(kaapi_cpuset_t)>=sizeof(kaapi_bitmap_value_t)
    */
    kaapi_assert_debug( sizeof(kaapi_cpuset_t)>=sizeof(kaapi_bitmap_value_t) )
    if (attr->init & KAAPIC_FOREACH_ATTR_INIT_CPUSET)
    {
      kaapi_bitmap_value_t all_cpus;
      kaapi_bitmap_value_clear( &gwork->wi.threadset );
      memcpy( &gwork->wi.threadset, &attr->threadset, sizeof(kaapi_bitmap_value_t) );
      
      /* mask the threadset with the available cpus in case if the user has set more bits */
      kaapi_bitmap_value_set_low_bits(&all_cpus, concurrency);
      kaapi_bitmap_value_and(&gwork->wi.threadset, &all_cpus);
      
      /* add localtid in the threadset */
      kaapi_bitmap_value_set( &gwork->wi.threadset, localtid );
    }
    else
    { /* accept requests from all kprocessors */
      kaapi_bitmap_value_set_low_bits(&gwork->wi.threadset, concurrency);
    }
    int cnt = kaapi_bitmap_value_count( &gwork->wi.threadset );
    if (cnt < gwork->wi.nthreads)
      gwork->wi.nthreads = cnt;
  }
  
  gwork->body_f    = body_f;
  gwork->body_args = body_args;

  /* initialize work distribution
  */
  _kaapic_foreach_initwa(
      &gwork->wa,
      gwork->wi.policy,
      localtid,
      gwork->wi.nthreads,
      &gwork->wi.threadset,
      first, last
  );
  
  /* reset the work remain/wokerdone field */
  KAAPI_ATOMIC_WRITE(&gwork->workremain, last - first);
  KAAPI_ATOMIC_WRITE(&gwork->workerdone, 0);
  KAAPI_ATOMIC_WRITE(&gwork->nthreads, 0);

  /* Initialize all the localwork to empty range (but lock is initialized !)
  */
  for (i=0; i<concurrency; ++i)
  {
    lwork = &gwork->lwork[i];
    /* initialize the lwork */
#if defined(KAAPIC_USE_KPROC_LOCK)
    kaapi_workqueue_init_with_lock(
      &lwork->cr,
      0, 0,
      &kaapi_all_kprocessors[i]->lock
    );
#else
    kaapi_atomic_initlock(&lwork->lock);
    kaapi_workqueue_init_with_lock(
      &lwork->cr, 
      0, 0,
      &lwork->lock
    );
#endif
    lwork->context  = 0;
    lwork->global   = gwork;
    lwork->workdone = 0;
    lwork->tid      = i;
    lwork->init     = 0;
  }
  
  return gwork;
}  



/* Initialize the local work + the global work
   Return a pointer to the local work to execute. 
   Do not require that reset is protected:
   - steal cannot occurs on local work while ->init ==0. 
*/
int kaapic_foreach_local_workinit
(
  kaapic_local_work_t*    lwork,
  kaapi_workqueue_index_t first,
  kaapi_workqueue_index_t last
)
{  
  kaapi_assert_debug(lwork !=0);
  kaapi_assert_debug(lwork->workdone == 0);
  kaapi_assert_debug(lwork->init == 0);

  kaapi_workqueue_reset(
    &lwork->cr, 
    first,
    last
  );

  lwork->init        = 1;

  return 0;
}  


/* task splitter: interface between scheduler and kaapic work (here foreach loop).
   This function is called when an idle kproc emits a request in order to extract work from
   the current task (foreach adaptive task).
   The method applies a filter to received requests :
   - if request->ident is not in the threadset, then the request is rejected
   - if the number of threads that have steal is greather than gwork->wi.nthreads, the request
   is rejected
   
   Then, depending of the policy attribut pass to the invocation of the foreach :
   - if the policy is KAAPIC_FOREACH_SCHED_STEAL: 
     The non rejected requests are aggregated and thieves steal a range of size
     nrequests/(1+nrequests).
   - if the policy is KAAPIC_FOREACH_SCHED_DISTRIBUTE:
     The non rejected requests are served using either the mapping in gwork->wa.map[tid] to
     give back their reserved field (if cpuset attribut was given in foreach), else the first
     non 0 bits in gwork->wa.map are used to give back work in gwork->wa.startindex[ithbits] to
     gwork->wa.startindex[ithbits+1].
  
  If the explanation is not clear, please read first the API-C documentation or contact 
  the online documentation (if you have a premium contract).
*/
static int _kaapic_split_task
(
  struct kaapi_task_t*                 victim_task,
  void*                                arg,
  struct kaapi_listrequest_t*          lr,
  struct kaapi_listrequest_iterator_t* lri
)
{
  /* call by a thief to steal work */
  int tid;
  kaapic_local_work_t* const lwork  = (kaapic_local_work_t*)arg;
  kaapic_global_work_t* const gwork = lwork->global;
  kaapi_bitmap_value_t keeprequests;
  kaapi_bitmap_value_t rootrequests;
  kaapi_workqueue_index_t first, last;
  kaapi_request_t* req;
  kaapic_local_work_t* tw;
  kaapi_workqueue_index_t range_size, unit_size;

  /* no attached body: return (typically case of OpenMP)*/
  if (gwork->body_f ==0)
    return 0;
  
  /* filter the requests that can be served.
  */
  kaapi_listrequest_iterator_t cpy_lri = *lri;
  kaapi_bitmap_value_clear( &keeprequests );
  kaapi_bitmap_value_clear( &rootrequests );
  for (; !kaapi_listrequest_iterator_empty(&cpy_lri); ) 
  {
    kaapi_request_t* req = kaapi_listrequest_iterator_get(lr, &cpy_lri);
    if (kaapi_bitmap_value_get( &gwork->wi.threadset, req->ident) !=0)
    {
      if (gwork->wi.policy == KAAPIC_FOREACH_SCHED_DISTRIBUTE)
      { /* build the set of requests to be served from reserved bloc */
        int pos = gwork->wa.tid2pos[ req->ident ];
        if (pos != (uint8_t)-1)
          kaapi_bitmap_value_set( &rootrequests, req->ident );
      }
      kaapi_bitmap_value_set( &keeprequests, req->ident );
    }
    kaapi_listrequest_iterator_next(lr, &cpy_lri);
  }

  if (kaapi_bitmap_value_empty(&keeprequests))
    return 0;
    
  if (gwork->wi.policy == KAAPIC_FOREACH_SCHED_DISTRIBUTE)
  {
    /* reply for first time to thread with reserved request (if !=0) */
    while ((tid = kaapi_bitmap_value_first1_and_zero(&rootrequests)) != 0)
    {
      /* because 1+tid is returned */
      --tid;

      /* use only get: the bitmap gwork->wa.map was updated atomically above */
      if ( kaapic_global_work_pop( gwork, tid, &first, &last,
                                  (gwork->wi.flags & KAAPIC_FOREACH_ATTR_INIT_CPUSET)!= 0 ? 0 : 1)
         )
      {
        req = &kaapi_global_requests_list[tid];
        KAAPI_ATOMIC_INCR(&gwork->workerdone);
        tw = &gwork->lwork[tid];
        kaapic_foreach_local_workinit(tw, first, last);

        kaapi_task_init_with_flag(
          kaapi_request_toptask(req), 
          (kaapi_task_body_t)_kaapic_thief_entrypoint, 
          tw,
          KAAPI_TASK_UNSTEALABLE
        );
  #if 1 /* comment this line if you do not want thief to be thief by other */
        kaapi_request_pushtask_adaptive_tail( 
          req, 
          victim_task,
          _kaapic_split_task 
        );
  #else // unstealable task 
        kaapi_request_pushtask(
          req,
          victim_task
        );
  #endif
        kaapi_request_committask(req);
        
        /* unset all thieves for which a reply has been sent */
        kaapi_listrequest_iterator_unset_at( lri, tid );

        /* also reset the entries in keeprequests */
        kaapi_bitmap_value_unset( &keeprequests, tid );
      } /* else: no work to done, will be reply failed by the runtime */
    }

    if (kaapi_bitmap_value_empty(&keeprequests))
      return 0;
  }

  /* here it case of SCHED_STEAL policy or cases where SCHED_DISTRIBUTE was not able to
     reply a reserved bloc for the kprocessors
  */
  /* because object lwork may exist without thread, test if initialized */
  if (lwork->init == 0)
    return 0;

  /* simple test to refuse requests : no enough work !
  */
  range_size = kaapi_workqueue_size(&lwork->cr);
  if (range_size < gwork->wi.rep.li.par_grain)
    return 0;

  /* how much per non root req */
  int nreq = kaapi_bitmap_value_count(&keeprequests);
  int addednthreads = nreq;
  if (gwork->wi.flags & KAAPIC_FOREACH_ATTR_INIT_NTHREADS)
  {
    int newval = KAAPI_ATOMIC_ADD(&gwork->nthreads, addednthreads);
    if (newval > gwork->wi.nthreads)
    {
      int toadd = gwork->wi.nthreads - (newval - addednthreads);
      if (toadd > addednthreads)
      {
        nreq = toadd;
      }
    }
  }
  
  unit_size = range_size / (nreq + 1);
  if (unit_size < gwork->wi.rep.li.par_grain)
  {
    nreq = (int)((range_size / gwork->wi.rep.li.par_grain) - 1);
    unit_size = gwork->wi.rep.li.par_grain;
    if (nreq ==0)
      return 0;
  }

  if (gwork->wi.flags & KAAPIC_FOREACH_ATTR_INIT_NTHREADS)
  {
    /* update the number of added nthreads to gwork */
    if (addednthreads != nreq)
      KAAPI_ATOMIC_ADD(&gwork->nthreads, nreq-addednthreads);
  }

#if defined(KAAPIC_USE_KPROC_LOCK)
  kaapi_assert_debug( lwork->cr.lock 
    == &kaapi_get_current_processor()->victim_kproc->lock );
  kaapi_assert_debug( kaapi_atomic_assertlocked(lwork->cr.lock) );
#else
  _kaapi_workqueue_lock(&lwork->cr);
#endif

  if (0 !=kaapi_workqueue_steal(&lwork->cr, &first, &last, nreq * unit_size))
  {
#if defined(KAAPIC_USE_KPROC_LOCK)
#else
    _kaapi_workqueue_unlock(&lwork->cr);
#endif
    KAAPI_LOCAL_LOG( printf("[%i]:: [_kaapic_split_task] fail steal %i: [%i, %i), victim queue: [%i, %i)\n", kaapi_get_self_kid(), lwork->tid, first, last, kaapi_workqueue_range_begin(&lwork->cr), kaapi_workqueue_range_end(&lwork->cr) ); fflush(stdout); )

    /* fail: update the number of threads attached to gwork */
    if (gwork->wi.flags & KAAPIC_FOREACH_ATTR_INIT_NTHREADS)
      KAAPI_ATOMIC_SUB(&gwork->nthreads, nreq);
    return 0;
  }
#if defined(KAAPIC_USE_KPROC_LOCK)
#else
  _kaapi_workqueue_unlock(&lwork->cr);
#endif
  KAAPI_LOCAL_LOG( printf("[%i]:: [_kaapic_split_task] success steal %i: [%i, %i)\n", kaapi_get_self_kid(), lwork->tid, first, last); fflush(stdout); )

  kaapi_assert_debug(first < last);
  kaapi_assert_debug(unit_size*nreq == last-first);  
#if defined(KAAPI_DEBUG)
  int sfirst __attribute__((unused)) = first;
  int slast __attribute__((unused))  = last;
#endif

  while (((tid = kaapi_bitmap_value_first1_and_zero(&keeprequests)) != 0) && (first<last))
  {
    --tid;
    req = &kaapi_global_requests_list[tid];
    tw = &gwork->lwork[tid];
    kaapic_foreach_local_workinit(tw, last-unit_size, last);

    kaapi_task_init_with_flag(
      kaapi_request_toptask(req), 
      (kaapi_task_body_t)_kaapic_thief_entrypoint, 
      tw,
      KAAPI_TASK_UNSTEALABLE
    );
#if 1 /* comment this line if you do not want thief to be thief by other */
    kaapi_request_pushtask_adaptive_tail( 
      req, 
      victim_task,
      _kaapic_split_task 
    );
#else // unstealable task 
    kaapi_request_pushtask(
      req,
      victim_task
    );
#endif
    kaapi_request_committask(req);
    
    /* mark request in lri set as replied */
    kaapi_listrequest_iterator_unset_at( lri, tid );
    last -= unit_size;
    kaapi_assert_debug(first <= last);
  }
  kaapi_assert_debug( last == first );

  return 0;
}


/* thief entrypoint, lwork is already initialized 
*/
static void _kaapic_thief_entrypoint(
  void*                 arg, 
  kaapi_thread_t*       thread,
  kaapi_task_t*         pc
)
{
  /* range to process */
  kaapi_workqueue_index_t i;
  kaapi_workqueue_index_t j;
  kaapi_workqueue_index_t lcounter =0; /* debug */

  kaapi_processor_t* kproc = kaapi_get_current_processor();

  KAAPI_EVENT_PUSH0(kproc, 0, KAAPI_EVT_FOREACH_BEG );

  /* process the work */
  kaapic_local_work_t* const lwork = (kaapic_local_work_t*)arg;
  kaapic_global_work_t* const gwork = lwork->global;

  /* asserts */
  kaapi_assert_debug(lwork->workdone == 0);
  kaapi_assert_debug(kaapi_get_self_kid() == lwork->tid);
#if defined(KAAPIC_USE_KPROC_LOCK)
  kaapi_assert_debug( &kproc->lock == lwork->cr.lock );
#else
#endif
  kaapi_assert_debug( kproc->kid == lwork->tid );

  KAAPI_LOCAL_LOG( printf("[%i]:: Entrypoint : begin: %i, end: %i\n", kaapi_get_self_kid(), kaapi_workqueue_range_begin(&lwork->cr), kaapi_workqueue_range_end(&lwork->cr)); fflush(stdout); )

  /* while there is sequential work to do in local work */
  while (kaapic_foreach_worknext(lwork, &i, &j) !=0)
  {
redo_local_work:
    kaapi_assert_debug( i < j );
    /* apply w->f on [i, j[ */
    gwork->body_f((int)i, (int)j, (int)lwork->tid, gwork->body_args);
    lwork->workdone += j-i;
    lcounter += j-i;
  }
  KAAPI_LOCAL_LOG( printf("[%i]:: Entrypoint: end loop, #iter: %i, %i\n", kaapi_get_self_kid(), lwork->workdone, lcounter ); fflush(stdout); )
  KAAPI_LOCAL_LOG( printf("[%i]:: Entrypoint: wq=[%i, %i]\n", kaapi_get_self_kid(), kaapi_workqueue_range_begin(&lwork->cr), kaapi_workqueue_range_end(&lwork->cr) ); fflush(stdout); )

  kaapi_assert_debug( kaapi_workqueue_isempty(&lwork->cr) );

  if (KAAPI_ATOMIC_SUB(&gwork->workremain, lwork->workdone) ==0) 
    goto return_label;

  kaapi_assert_debug(KAAPI_ATOMIC_READ(&gwork->workremain) >=0);
  lwork->workdone = 0;

  KAAPI_EVENT_PUSH0(kproc, 0, KAAPI_EVT_SCHED_IDLE_BEG );
  int retval = kaapic_foreach_globalwork_next( lwork, &i, &j );
  KAAPI_EVENT_PUSH0(kproc, 0, KAAPI_EVT_SCHED_IDLE_END );
  if (retval)
    goto redo_local_work;
  
return_label:
  KAAPI_EVENT_PUSH0(kproc, 0, KAAPI_EVT_FOREACH_END );
  KAAPI_LOCAL_LOG( printf("[%i]:: End Entrypoint : #workdone: %i\n", kaapi_get_self_kid(), lcounter ); fflush(stdout); )

  lwork->workdone = 0;
  lwork->init = 0;

  /* suppress lwork reference */
  kaapi_task_unset_splittable(pc);
  kaapi_synchronize_steal_thread(kproc->thread);

  KAAPI_ATOMIC_DECR(&gwork->workerdone);
}




/* Initialize the local work + the global work
   Return a pointer to the local work to execute. 
*/
kaapic_local_work_t* kaapic_foreach_workinit
(
  kaapi_thread_context_t* self_thread,
  kaapi_workqueue_index_t first, 
  kaapi_workqueue_index_t last,
  const kaapic_foreach_attr_t*  attr,
  kaapic_foreach_body_t   body_f,
  kaapic_body_arg_t*      body_args
)
{
  kaapic_global_work_t* gwork;
  kaapic_local_work_t*  lwork = 0;

  if (last <= first) return 0;

  const int self_tid = self_thread->stack.proc->kid;

  gwork = kaapic_foreach_global_workinit(
      self_thread, 
      first, 
      last, 
      attr, 
      body_f,
      body_args
  );
  if (gwork ==0)
    return 0;
  kaapi_assert_debug(gwork !=0);
  
  /* initialize the local workqueue with the first poped state */
  if (kaapic_global_work_pop( gwork, self_tid, &first, &last, 0))
  {    
    /* */
    lwork = &gwork->lwork[self_tid];
    kaapic_foreach_local_workinit(
                    lwork,
                    first,
                    last
    );
    /* from here lwork is visible to be steal, but it cannot until
       the adaptive task is spawned 
    */
  }
  kaapi_assert_debug(lwork !=0);

  /* start adaptive region */
  lwork->context = kaapi_task_begin_adaptive(
     kaapi_threadcontext2thread(self_thread), 
     KAAPI_SC_CONCURRENT | KAAPI_SC_NOPREEMPTION,
     _kaapic_split_task,
     lwork
  );

  /* begin a parallel region */
  if (kaapic_do_parallel)
    kaapic_begin_parallel(KAAPIC_FLAG_DEFAULT);

  return lwork;
}  


int kaapic_foreach_local_workend(
  kaapi_thread_context_t* self_thread,
  kaapic_local_work_t*    lwork
)
{
  lwork->init = 0;
  
  /* exec: task and wait end of adaptive task */
  kaapi_sched_sync_(self_thread);

#if defined(KAAPIC_USE_KPROC_LOCK)
#else
  kaapi_atomic_destroylock(&lwork->lock);
#endif
  return 0;
}


/* Symetric of kaapic_foreach_workinit */
int kaapic_foreach_workend
(
  kaapi_thread_context_t* self_thread,
  kaapic_local_work_t*    lwork
)
{
  kaapic_global_work_t* gwork = lwork->global;
  kaapi_assert_debug(gwork != 0);
  
  /* push task to wait for thieves */
  kaapi_task_end_adaptive(
    kaapi_threadcontext2thread(self_thread), 
    lwork->context
  );
  
  /* exec: task and wait end of adaptive task */
  kaapi_sched_sync_(self_thread);

  if (kaapic_do_parallel) 
    kaapic_end_parallel(KAAPI_SCHEDFLAG_DEFAULT);
  
/* WARNING: adaptive tasks must ensure terminaison ! */ /* wait worker */
  while (KAAPI_ATOMIC_READ(&gwork->workerdone) >0)
    kaapi_slowdown_cpu();

  /* after this instruction: global + local work disapear */
  kaapi_thread_pop_frame_( self_thread );

#if defined(KAAPIC_USE_KPROC_LOCK)
#else
  kaapi_atomic_destroylock(&lwork->cr.lock);
#endif  

  return 0;
}


/* 
  Return !=0 iff first and last have been filled for the next piece
  of work to execute.
  The function try to steal from registered lwork in the global work.
  The workqueue is fill by poped range.
  In case of data distribution attribut, the localworkqueue_t structure
  if filled to the poped range and the returned first,last is the biggest
  contiguous range of iteration.
*/
static int kaapic_foreach_globalwork_next(
  kaapic_local_work_t*     lwork,
  kaapi_workqueue_index_t* first,
  kaapi_workqueue_index_t* last
)
{
  kaapic_global_work_t* gwork = lwork->global;
  kaapi_processor_t* kproc = kaapi_get_current_processor();

  while (KAAPI_ATOMIC_READ(&gwork->workremain) !=0)
  {
    kaapi_assert_debug(KAAPI_ATOMIC_READ(&gwork->workremain) >=0);

    if (kaapic_global_work_steal(
          gwork,
          kproc,
          first,last)
       )
    {
      /* return directly the range if less than sequential grain */
      if (*last - *first <= gwork->wi.rep.li.seq_grain)
      {
        KAAPI_LOCAL_LOG( printf("[%i]:: [kaapic_foreach_globalwork_next] steal (<grain) [%i,%i)\n", kaapi_get_self_kid(), *first, *last); fflush(stdout); )
        return 1;
      }

      /* refill the global work data structure without seq_grain size range for next pop */
      kaapi_workqueue_index_t beg = *first+gwork->wi.rep.li.seq_grain;
      kaapi_workqueue_index_t end = *last;
      _kaapi_workqueue_lock( &lwork->cr );
      kaapic_foreach_local_workinit( 
          lwork,
          beg,
          end
      );
      _kaapi_workqueue_unlock( &lwork->cr );
      KAAPI_LOCAL_LOG( printf("[%i]:: [kaapic_foreach_globalwork_next] steal [%i,%i), init queue: [%i,%i) / [%i,%i), get [%i,%i)\n", kaapi_get_self_kid(), *first, *last, beg, end, kaapi_workqueue_range_begin(&lwork->cr), kaapi_workqueue_range_end(&lwork->cr), *first, beg); fflush(stdout); )
      *last = beg;
      return 1;
    }
    kaapi_slowdown_cpu();
  }
  return 0; /* means global is terminated */
}


/* 
  Return !=0 iff first and last have been filled for the next piece
  of work to execute.
*/
int kaapic_foreach_worknext(
  kaapic_local_work_t*     lwork,
  kaapi_workqueue_index_t* first,
  kaapi_workqueue_index_t* last
)
{
  int retval;
  kaapic_global_work_t* gwork = lwork->global;
  long sgrain = gwork->wi.rep.li.seq_grain;

  KAAPI_EVENT_PUSH0(kaapi_get_current_processor(), 0, KAAPI_EVT_SCHED_IDLE_BEG );
  
  if (kaapi_workqueue_pop(&lwork->cr, first, last, sgrain) == 0)
  {
    retval = 1;
    kaapi_assert_debug( *first < *last );
    goto return_value;
  }

  kaapi_assert_debug( kaapi_workqueue_isempty(&lwork->cr) );
  lwork->init = 0;
  retval = 0;
  
return_value:
  KAAPI_EVENT_PUSH0(kaapi_get_current_processor(), 0, KAAPI_EVT_SCHED_IDLE_END );
  return retval;
}


/* exported foreach interface */
int kaapic_foreach_common
(
  kaapi_workqueue_index_t      first,
  kaapi_workqueue_index_t      last,
  const kaapic_foreach_attr_t* attr,
  kaapic_foreach_body_t        body_f,
  kaapic_body_arg_t*           body_args
)
{
  kaapic_global_work_t* gwork;
  kaapic_local_work_t*  lwork;
  
  kaapi_workqueue_index_t lcounter =0;
  
  if (last <= first) 
    return 0;

  /* is_format true if called from kaapif_foreach_with_format */
  kaapi_processor_t* const kproc = kaapi_get_current_processor();
  kaapi_thread_context_t* const self_thread = kproc->thread;
  const int self_tid = kproc->kid;

  KAAPI_EVENT_PUSH0(kproc, 0, KAAPI_EVT_FOREACH_BEG );

  /* initalize the foreach global and local work, returns the local work pointer or 0 */
  lwork = kaapic_foreach_workinit(
      self_thread,
      first,
      last,
      attr, 
      body_f,
      body_args
  );
  kaapi_assert_debug(lwork !=0);
  
  gwork = lwork->global;
  
  KAAPI_LOCAL_LOG( printf("[%i]:: master : begin: %i, end: %i\n", kaapi_get_self_kid(), kaapi_workqueue_range_begin(&lwork->cr), kaapi_workqueue_range_end(&lwork->cr)); fflush(stdout); )
  /* while there is sequential work to do in local work */
  while (kaapic_foreach_worknext(lwork, &first, &last) !=0)
  {
redo_local_work:
    kaapi_assert_debug( first < last );
    /* apply w->f on [i, j[ */
    body_f((int)first, (int)last, (int)self_tid, body_args);
    lwork->workdone += last-first;
    lcounter += last-first;
  }
  kaapi_assert_debug( kaapi_workqueue_isempty(&lwork->cr) );
  KAAPI_LOCAL_LOG( printf("[%i]:: master: end loop, #iter: %i\n", kaapi_get_self_kid(), lwork->workdone ); fflush(stdout); )

  /* update workload information and return if finished */
  if (KAAPI_ATOMIC_SUB(&gwork->workremain, lwork->workdone) ==0) 
    goto return_label;

  kaapi_assert_debug(KAAPI_ATOMIC_READ(&gwork->workremain) >=0);
  
  /* reset local workdone before next range */
  lwork->workdone = 0;

  KAAPI_EVENT_PUSH0( kproc, 0, KAAPI_EVT_SCHED_IDLE_BEG );
  int retval = kaapic_foreach_globalwork_next( lwork, &first, &last );
  KAAPI_EVENT_PUSH0( kproc, 0, KAAPI_EVT_SCHED_IDLE_END );
  if (retval)
    goto redo_local_work;
 
return_label:
  KAAPI_EVENT_PUSH0(kproc, 0, KAAPI_EVT_FOREACH_END );
  KAAPI_LOCAL_LOG( printf("[%i]:: End master : #workdone: %i\n", kaapi_get_self_kid(), lcounter ); fflush(stdout); )

  lwork->workdone = 0;

  kaapic_foreach_workend( self_thread, lwork );

  return 0;
}
