/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** francois.broquedis@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <errno.h>

#include "kaapi_impl.h"
#include "kaapic_impl.h"

unsigned long komp_env_nthreads          = 0;
unsigned long komp_env_nested            = 0;
unsigned long komp_env_stacksize         = 0;
char *        komp_env_cpu_affinity      = NULL;

unsigned long omp_max_active_levels      = 0;

/* Parse the OMP_STACKSIZE environment varible.  Return true if one was
   present and it was successfully parsed.  */

static bool
parse_stacksize (const char *name, unsigned long *pvalue)
{
  char *env, *end;
  unsigned long value, shift = 10;

  env = getenv (name);
  if (env == NULL)
    return false;

  while (isspace ((unsigned char) *env))
    ++env;
  if (*env == '\0')
    goto invalid;

  errno = 0;
  value = strtoul (env, &end, 10);
  if (errno)
    goto invalid;

  while (isspace ((unsigned char) *end))
    ++end;
  if (*end != '\0')
    {
      switch (tolower ((unsigned char) *end))
	{
	case 'b':
	  shift = 0;
	  break;
	case 'k':
	  break;
	case 'm':
	  shift = 20;
	  break;
	case 'g':
	  shift = 30;
	  break;
	default:
	  goto invalid;
	}
      ++end;
      while (isspace ((unsigned char) *end))
	++end;
      if (*end != '\0')
	goto invalid;
    }

  if (((value << shift) >> shift) != value)
    goto invalid;

  *pvalue = value << shift;
  return true;

 invalid:
  fprintf (stderr, "Invalid value for environment variable %s", name);
  return false;
}

/* Parse an unsigned long environment varible.  Return true if one was
   present and it was successfully parsed.  */

static bool
parse_unsigned_long (const char *name, unsigned long *pvalue)
{
  char *env, *end;
  unsigned long value;

  env = getenv (name);
  if (env == NULL)
    return false;

  while (isspace ((unsigned char) *env))
    ++env;
  if (*env == '\0')
    goto invalid;

  errno = 0;
  value = strtoul (env, &end, 10);
  if (errno || (long) value <= 0)
    goto invalid;

  while (isspace ((unsigned char) *end))
    ++end;
  if (*end != '\0')
    goto invalid;

  *pvalue = value;
  return true;

 invalid:
  fprintf (stderr, "Invalid value for environment variable %s", name);
  return false;
}

#define __append_string_with_format(kaapi_cpuset_string, format, ...) \
do {                                                                  \
  char tmp[16];                                                       \
  sprintf (tmp, format, __VA_ARGS__);                                 \
  written_chars += strlen (tmp);				      \
  if (written_chars >= string_len) {				      \
    string_len *= 2;						      \
    kaapi_cpuset_string = realloc (kaapi_cpuset_string, string_len);  \
  }								      \
  strcat (kaapi_cpuset_string, tmp);                                  \
 } while (0)


/* This function parses the content of the GOMP_CPU_AFFINITY 
   environment variable to generate something that fits the semantic 
   of the KAAPI_CPUSET environment variable. */
static void
komp_parse_cpu_affinity (void)
{
  char *cpu_affinity = getenv ("GOMP_CPU_AFFINITY");
  if (cpu_affinity == NULL)
    return;
  
  int string_len = 512;
  int written_chars = 0;
  komp_env_cpu_affinity = calloc (512, sizeof (char));
  
  char *cpuset_string = NULL;  
  while ((cpuset_string = strsep (&cpu_affinity, " \t")) != NULL)
  {
    char *first_cpu = strsep (&cpuset_string, "-");
    char *last_cpu = strsep (&cpuset_string, ":");
    char *stride = cpuset_string;

    if (stride != NULL)
    {
      int begin = atoi (first_cpu);
      int end = atoi (last_cpu);
      int hop = atoi (stride);
      
      int i;
      for (i = begin; i < end; i += hop)
        __append_string_with_format (komp_env_cpu_affinity, "%d,", i); 

    } else {
      if (last_cpu != NULL)
        __append_string_with_format (komp_env_cpu_affinity, "%s:%s,", first_cpu, last_cpu); 
      else 
        __append_string_with_format (komp_env_cpu_affinity, "%s,", first_cpu); 
    }
  }
  komp_env_cpu_affinity[strlen (komp_env_cpu_affinity) - 1] = '\0';
  setenv ("KAAPI_CPUSET", komp_env_cpu_affinity, 1);
}

static inline const char *
komp_env_msg (unsigned long var, unsigned long value) {
  if (var == value)
    return "undefined: will use OpenMP default";
  return "user-defined";
}

/* Display runtime-related information as a banner printed before
   running the OpenMP application. */
static void
komp_display_env (void) {
  fprintf (stderr, "[libKOMP] OMP_NUM_THREADS\t=\t%lu\t(%s)\n", 
	   komp_env_nthreads, komp_env_msg(komp_env_nthreads, 0));
  fprintf (stderr, "[libKOMP] OMP_NESTED\t\t=\t%lu\t(%s)\n", 
	   komp_env_nested, komp_env_msg(komp_env_nested, 0));
  fprintf (stderr, "[libKOMP] OMP_STACKSIZE\t\t=\t%lu\t(%s)\n", 
	   komp_env_stacksize, komp_env_msg(komp_env_stacksize, 0));
  fprintf (stderr, "[libKOMP] OMP_MAX_ACTIVE_LEVELS\t=\t%lu\t(%s)\n", 
	   omp_max_active_levels, komp_env_msg(omp_max_active_levels, 0));
  fprintf (stderr, "[libKOMP] GOMP_CPU_AFFINITY\t=\t%s\t(%s)\n", 
	   komp_env_cpu_affinity, komp_env_msg((unsigned long)komp_env_cpu_affinity, 0));
}

/* Parse the OMP_DIPLAY_ENV environment variable.
   This feature comes from OpenMP 4 RC2.

   As described in section 4.12, OMP_DISPLAY_ENV can be set to one of
   these values: 
   true | false | verbose

   and display runtime-related information accordingly (ICVs, version
   number, ...).

   Note that, as opposed to the current version of the 4RC2 specs,
   lowercase characters have been chosen to fit the style of other
   OpenMP variables like OMP_NESTED.
 */
static void
komp_parse_display_env (void) {
  char *env = getenv ("OMP_DISPLAY_ENV");
  if (env == NULL)
    return;

  if (strcmp (env, "true") == 0) {
    komp_display_env ();
  } else if (strcmp (env, "verbose") == 0) {
    komp_display_env ();
  } else if (strcmp (env, "false") != 0) {
    fprintf (stderr, "[libKOMP] Wrong value for OMP_DISPLAY_ENV (%s)! Should be set to one of (true | false | verbose).\n", env);
  }
}

/* Parse the OMP_NESTED environment variable. */
static void
komp_parse_nested_env (void) {
  char *env = getenv ("OMP_NESTED");
  if (env == NULL)
    return;

  if (strcmp (env, "true") == 0) {
    komp_env_nested = 1;
  } else if (strcmp (env, "false") == 0) {
    komp_env_nested = 0;
  } else {
    fprintf (stderr, "[libKOMP] Wrong value for OMP_NESTED (%s)! Should be set to one of (true | false).\n", env);
  }
}

extern double kaapi_komp_start_parallel;
extern int kaapi_komp_start_parallel_count;
extern double kaapi_komp_end_parallel;
extern int kaapi_komp_end_parallel_count;

static void __attribute__ ((constructor))  
initialize_lib (void) 
{
  char *show_banner = getenv ("KAAPI_GOMP_SHOW_BANNER");
  if (show_banner && *show_banner) {
    fprintf(stderr, "Using " PACKAGE " as GNU OpenMP runtime\n");
  }
  
  /* Turn GOMP_CPU_AFFINITY into KAAPI_CPUSET. */
  komp_parse_cpu_affinity ();
  
  /* Turn OMP_NUM_THREADS into KAAPI_CPUCOUNT. */
  if (parse_unsigned_long ("OMP_NUM_THREADS", &komp_env_nthreads)) {
    setenv ("KAAPI_CPUCOUNT", getenv ("OMP_NUM_THREADS"), 1);
  }

  /* Turn GOMP_STACKSIZE and OMP_STACKSIZE into KAAPI_STACKSIZE. */
  if (parse_stacksize ("OMP_STACKSIZE", &komp_env_stacksize)
      || parse_stacksize ("GOMP_STACKSIZE", &komp_env_stacksize)) {
    char stackenv[64];
    sprintf (stackenv, "%lu", komp_env_stacksize);
    setenv ("KAAPI_STACKSIZE", stackenv, 1);
  }

  /* Parse OMP_MAX_ACTIVE_LEVELS. */
  parse_unsigned_long ("OMP_MAX_ACTIVE_LEVELS", &omp_max_active_levels);

  /* Parse OMP_NESTED. */
  komp_parse_nested_env ();

  /* Parse OMP_DISPLAY_ENV. */
  komp_parse_display_env ();

  if (omp_max_active_levels == 0)
    omp_max_active_levels = KAAPI_MAX_RECCALL;

  kaapi_komp_start_parallel= 0;
  kaapi_komp_start_parallel_count= 0;
  kaapi_komp_end_parallel= 0;
  kaapi_komp_end_parallel_count= 0;

  kaapic_init (KAAPIC_START_ONLY_MAIN);
}


static void __attribute__ ((destructor))
finalize_lib (void)
{
  kaapic_finalize ();
#if KAAPI_KOMP_TRACE
  printf("KOMP: time start parallel: %f\n", 1e6*kaapi_komp_start_parallel/kaapi_komp_start_parallel_count); 
  printf("KOMP: time end parallel: %f\n", 1e6*kaapi_komp_end_parallel/kaapi_komp_end_parallel_count); 
#endif
}
