/**
 *
 * @file quark_impl.h
 *
 * Dynamic scheduler functions
 *
 * PLASMA is a software package provided by Univ. of Tennessee,
 * Univ. of California Berkeley and Univ. of Colorado Denver
 *
 ** xkaapi
 **
 ** Copyright 2011 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** jvlima@inf.ufrgs.br
 **
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 **
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 **
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 **
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 **
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
*/

#ifndef _KAAPI_QUARK_H_
#define _KAAPI_QUARK_H_

#include "kaapi_impl.h"

/* see api/quark for more information */
struct quark_s;

typedef void (*kaapi_quark_task_body_t)(struct quark_s *);

struct quark_s *kaapi_quark_get_current(void);

typedef struct quark_one_param_t {
  kaapi_access_t      addr;  /* .data used to store value */
  size_t              size;
  kaapi_access_mode_t mode;
} quark_one_param_t __attribute__((aligned(8)));

#define MAX_PARAMQUARK 32

/* XKaapi interface for Quark task with variable number of parameters */
typedef struct quark_task_s {
  kaapi_quark_task_body_t function;
  char*                 name;
  uintptr_t             callitwith_handle;
  uint32_t              scratchbit;
  uint32_t              nparam;    /* number of parameters */
  quark_one_param_t     param[MAX_PARAMQUARK];
} kaapi_quark_task_t;

/* Type for task sequences */
typedef struct Quark_sequence_s {
  int                     state_init;  /* see flag above */
  int                     save_state;  /* */
  kaapi_frame_t           save_fp;     /* saved fp of the current stack before sequence creation */
  kaapi_frame_tasklist_t  frame_tasklist;    /* */
  struct Quark_sequence_s* next;        /* next sequence */
} kaapi_quark_sequence_t;


/* opaque structure to user's quark program
 */
typedef struct quark_s {
  kaapi_thread_context_t* thread;
  kaapi_quark_task_t*     task;     /* the quark task = arg of Kaapi task */
  kaapi_quark_sequence_t* sequence;
} XKaapi_Quark;

#if defined(KAAPI_USE_KBLAS)

kaapi_quark_task_body_t kaapi_quark_task_get_body(kaapi_quark_task_body_t plasma, const uint8_t arch);

void kaapi_quark_task_set_arch(kaapi_quark_task_body_t plasma, const uint8_t arch);

uint8_t kaapi_quark_task_get_arch(kaapi_quark_task_body_t plasma);

kaapi_quark_task_body_t kaapi_quark_task_get_wrapper(kaapi_quark_task_body_t plasma);

void kaapi_quark_task_format_register(const char* name, kaapi_quark_task_body_t plasma,
                                      kaapi_quark_task_body_t wrapper, kaapi_quark_task_body_t wrapper_wh, kaapi_quark_task_body_t body,
                                      kaapi_quark_task_body_t wrapper_cuda, kaapi_quark_task_body_t wrapper_cuda_wh, kaapi_quark_task_body_t body_cuda );

void kaapi_quark_task_enable(void);
void kaapi_quark_task_disable(void);

#define KAAPI_QUARK_WRAPPER_FUNCTION_NAME( quarkfunction )      kaapi_wrapper_quark_function_##quarkfunction
#define KAAPI_QUARK_WRAPPER_FUNCTION_WH_NAME( quarkfunction )   kaapi_wrapper_wh_quark_function_##quarkfunction

#define KAAPI_QUARK_REGISTER_WRAPPER_FUNCTION( quarkfunction ) \
  static inline void kaapi_wrapper_quark_function_##quarkfunction ( void* a, kaapi_thread_t* thread, kaapi_task_t* task ) \
  { \
    kaapi_quark_task_t* arg  = (kaapi_quark_task_t*)a; \
    kaapi_processor_t* kproc = kaapi_get_current_processor(); \
    XKaapi_Quark* myquark    = kaapi_quark_get_current(); \
    myquark->thread          = kproc->thread; \
    myquark->task            = arg; \
    arg->callitwith_handle   = 0; \
    kaapi_quark_task_body_t body  = kaapi_quark_task_get_body(arg->function, KAAPI_PROC_TYPE_CPU); \
    body( myquark ); \
  } \
  static inline void kaapi_wrapper_wh_quark_function_##quarkfunction ( void* a, kaapi_thread_t* thread, kaapi_task_t* task ) \
  { \
    kaapi_quark_task_t* arg  = (kaapi_quark_task_t*)a; \
    kaapi_processor_t* kproc = kaapi_get_current_processor(); \
    XKaapi_Quark* myquark    = kaapi_quark_get_current(); \
    myquark->thread          = kproc->thread; \
    myquark->task            = arg; \
    arg->callitwith_handle   = 1; \
    kaapi_quark_task_body_t body  = kaapi_quark_task_get_body(arg->function, KAAPI_PROC_TYPE_CPU); \
    body( myquark ); \
  }

#if defined(KAAPI_USE_CUDA)

#define KAAPI_QUARK_WRAPPER_CUDA_FUNCTION_NAME( quarkfunction )      kaapi_wrapper_cuda_quark_function_##quarkfunction
#define KAAPI_QUARK_WRAPPER_CUDA_FUNCTION_WH_NAME( quarkfunction )   kaapi_wrapper_cuda_wh_quark_function_##quarkfunction

#define KAAPI_QUARK_REGISTER_WRAPPER_CUDA_FUNCTION( quarkfunction ) \
  static inline void kaapi_wrapper_cuda_quark_function_##quarkfunction ( void* a, kaapi_gpustream_t stream ) \
  { \
    kaapi_quark_task_t* arg  = (kaapi_quark_task_t*)a; \
    kaapi_processor_t* kproc = kaapi_get_current_processor(); \
    XKaapi_Quark* myquark    = kaapi_quark_get_current(); \
    myquark->thread          = kproc->thread; \
    myquark->task            = arg; \
    arg->callitwith_handle   = 0; \
    kaapi_quark_task_body_t body  = kaapi_quark_task_get_body(arg->function, KAAPI_PROC_TYPE_CUDA); \
    body( myquark ); \
  } \
  static inline void kaapi_wrapper_cuda_wh_quark_function_##quarkfunction ( void* a, kaapi_gpustream_t stream ) \
  { \
    kaapi_quark_task_t* arg  = (kaapi_quark_task_t*)a; \
    kaapi_processor_t* kproc = kaapi_get_current_processor(); \
    XKaapi_Quark* myquark    = kaapi_quark_get_current(); \
    myquark->thread          = kproc->thread; \
    myquark->task            = arg; \
    arg->callitwith_handle   = 1; \
    kaapi_quark_task_body_t body  = kaapi_quark_task_get_body(arg->function, KAAPI_PROC_TYPE_CUDA); \
    body( myquark ); \
  }

#else /* KAAPI_USE_CUDA */

#define KAAPI_QUARK_WRAPPER_CUDA_FUNCTION_NAME( quarkfunction )      (0)
#define KAAPI_QUARK_WRAPPER_CUDA_FUNCTION_WH_NAME( quarkfunction )   (0)
#define KAAPI_QUARK_REGISTER_WRAPPER_CUDA_FUNCTION( quarkfunction )

#endif /* KAAPI_USE_CUDA */

#endif /* KAAPI_USE_KBLAS */

#endif /* _KAAPI_QUARK_H_ */