/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** Thierry Gautier, thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
/* Open a set of event files
*/
#include "kaapi_impl.h"
#include "kaapi_trace_simulator.h"
#include <assert.h>
#include <vector>
#include <float.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

/** see kaapi_event.h for coding of event name */
#if (__KAAPI_TRACE_VERSION__ >= 1)
static const char* kaapi_event_name[] 
__attribute__((unused)) 
= {
/* 0 */  "K-ProcStart",
/* 1 */  "K-ProcEnd",
/* 2 */  "TaskBegin",
/* 3 */  "TaskEnd",
/* 4 */  "BeginFrame",
/* 5 */  "EndFrame",
/* 6 */  "StaticScheduleBeg",
/* 7 */  "StaticScheduleEnd",
/* 8 */  "StaticTaskBeg",
/* 9 */  "StaticTaskEnd",
/*10 */  "IdleBeg",
/*11 */  "IdleEnd",
/*12 */  "SuspendBeg",
/*13 */  "SuspendEnd",
/*14 */  "PostSuspend",
/*15 */  "PostWaitBeg",
/*16 */  "PostWaitEnd",
/*17 */  "RequestBeg",
/*18 */  "RequestEnd",
/*19 */  "StealOp",
/*20 */  "SendReply",
/*21 */  "RecvReply",
/*22 */  "*",
/*23 */  "*",
/*24 */  "*",
/*25 */  "ForEachBeg",
/*26 */  "ForEachEnd",
/*27 */  "ForEachSteal",
/*28 */  "CudaCPUHtoDBeg",
/*29 */  "CudaCPUHtoDEnd",
/*30 */  "CudaCPUDtoHBeg",
/*31 */  "CudaCPUDtoHEnd",
/*32 */  "CudaSyncBeg",
/*33 */  "CudaSyncEnd",
/*34 */  "CudaAllocBeg",
/*35 */  "CudaAllocEnd",
/*36 */  "TaskMainBeg",
/*37 */  "TaskMainEnd",
/*38 */  "PerfCounterEvent"
};
#else
#error "Undefined trace version"
#endif

/* This stack size correspond to the initial maximal deepth of tasks supported
   by the simulation
*/
#define TASK_STACKSIZE 1024

struct Simulator;

struct ProcessorState {
  Simulator* sim;
  int      active;            /* 1 iff active, 0 iff idle */
  uint64_t taskcount;
  uint64_t stealcount;        /* successfull steal operation */
  uint64_t tactive;           /* time where is active  */
  uint64_t tidle;             /* time where is idle */
  uint64_t tcompute;          /* time where is passed in task */
  uint64_t lastdate;          /* last date when activity state changed */
  uint64_t lasttactive;       /* time where is active  */
  uint64_t lasttidle;         /* time where is idle */
  uint64_t firstdate;         /* start time */
  uint64_t finaldate;         /* final time */
  uint64_t tmax;              /* statistics about task execution unit== second*/  
  uint64_t tmin;
  uint64_t tmain;
  double   tsum;              /* sum of execution time */
  double   t2;                /* sum t*t */
  int      sp;                /* stack pointer */
  uint64_t* stacktask;        /* stack of start time for each stack: unit= nano second */
  int      stackdeepth;       /* maximal number of deepth in the stack */
  kaapi_perf_counter_t	   perfctr[2][KAAPI_PERF_ID_MAX];
};

static void processor_simulate_event(
  ProcessorState& p,
  const kaapi_event_t* event
);

/* state of the simulator */
struct Simulator {
  FileSet*                    fds;
  int                         discretetime;  /* !=0 if discrete time=>report events at their date */
  int                         flag_taskdump;
  int                         fd_taskdump;
  uint64_t                    hres;  /* in ns */
  uint64_t                    tmin;  /* in s */
  uint64_t                    tmax;  /* in s */
  uint64_t                    tsim;  /* in ns, start from 0 (offset with epoch == tmin) */
  std::vector<ProcessorState> procs;
  double*                     efficiencies;
  uint64_t*                   count_stealok;
  uint64_t*                   count_task;
};

/* Return the efficiency of the processor at the given date 
   tsim is the next sampling time.
*/
static double ProcessorEfficiencyAt( ProcessorState& proc, uint64_t tsim )
{
  uint64_t tidle;
  uint64_t tactive;
  uint64_t diff_tidle;
  uint64_t diff_tactive;
  
  if (proc.lastdate ==0) return 0;

  assert(proc.lastdate <= tsim);
  
  tidle   = proc.tidle;
  tactive = proc.tactive;

  /* update with state between lastdate and tsim */
  if (proc.active ==1) 
    tactive += (tsim - proc.lastdate);
  else
    tidle   += (tsim - proc.lastdate);

  /* monotone increasing functions bound by tsim ! */
  assert( tactive >= proc.tactive);
  assert( tidle >= proc.tidle);
  assert( tactive <= tsim );
  assert( tidle <= tsim );

  /* diff with previous sampling time at t0 */
  assert( tidle >= proc.lasttidle);
  diff_tidle  = tidle - proc.lasttidle;
  assert( tactive >= proc.lasttactive);
  diff_tactive = tactive - proc.lasttactive;
  
  /* */
  proc.lasttactive = tactive;
  proc.lasttidle   = tidle;
  
  if ((diff_tactive ==0) || (diff_tidle ==0))
    return 0;
  return (double)(diff_tactive) / (double)(diff_tactive+diff_tidle);
}


/*
*/
extern "C"
Simulator* OpenSimulator( FileSet* fds, double hres, int discrete, int flag_taskdump )
{
  int i, err;
  int count;
  
  Simulator* sim = new Simulator;
  if (sim ==0) return 0;
  sim->fds   = fds;
  sim->hres  = (uint64_t)(hres * 1e9); /* count in ns */
  sim->tsim  = 0;

  err = GetInterval(fds, &sim->tmin, &sim->tmax);
  if (err !=0)
    goto return_on_error;

  count = GetProcessorCount(fds);
  if (count <=0) 
    goto return_on_error;

  sim->procs.resize(count);
  sim->efficiencies  = (double*)malloc(sizeof(double) * count );
  sim->count_stealok = (uint64_t*)malloc(sizeof(uint64_t) * count );
  sim->count_task    = (uint64_t*)malloc(sizeof(uint64_t) * count );
  sim->discretetime  = (discrete !=0);
  sim->flag_taskdump = flag_taskdump;
  if (flag_taskdump)
  {
    /* Error using O_CREAT with no third parameter on Ubuntu */
    sim->fd_taskdump = open("task.time", O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR | S_IWUSR);
    if (sim->fd_taskdump == -1)
    {
      fprintf(stderr, "*** Cannot create 'task.time', error: '%s'\n", strerror(errno));
      fflush(stderr);
    }
    else {
      fprintf(stdout, "*** File 'task.time' created");
      fflush(stdout);
    }
  }
  else 
  {
    sim->fd_taskdump = -1;
  }
    
  for (i=0;i<count; ++i)
  {
    ProcessorState& p = sim->procs[i];
    p.sim        = sim;
    p.active     = 0;
    p.taskcount  = 0;
    p.stealcount = 0;
    p.tactive    = 0;
    p.tidle      = 0;
    p.lastdate   = 0;
    p.lasttactive= 0;
    p.lasttidle  = 0;
    p.firstdate  = 0;
    p.finaldate  = 0;
    p.tmax       = 0;
    p.tmin       = ~0UL;
    p.tmain      = 0;
    p.tsum       = 0;
    p.t2         = 0;
    p.sp         = -1; /* no task first ! */
    p.stackdeepth= TASK_STACKSIZE;
    p.stacktask  = (uint64_t*)calloc(sizeof(uint64_t), p.stackdeepth);
    if (p.stacktask ==0)
    {
      fprintf(stderr, "*** Error: cannot allocate memory\n");
      fflush(stderr);
      kaapi_abort();
    }
    bzero(&p.perfctr, sizeof(p.perfctr));
  }
  
  return sim;

return_on_error:
  delete sim;
  return 0;
}



extern "C"
double NextStepSimulator( Simulator* sim )
{
  const kaapi_event_t* event;
  uint64_t nextTsim =0; /* used for continuous mode */

  /* simulate until global date sim->tmin + sim->tsim + sim->hres */
  if (sim->discretetime ==0)
    nextTsim   = sim->tmin+sim->tsim+sim->hres;
    
  if (EmptyEvent(sim->fds))
  {
    return -1.0;
  }
  
  while (!EmptyEvent(sim->fds))
  {
    event = TopEvent(sim->fds);
    if (sim->discretetime ==0)
    {
      if (event->date >= nextTsim)
        break;
      
      processor_simulate_event( sim->procs[event->kid], event );
      
      NextEvent(sim->fds);
    }
    else
    {
      processor_simulate_event( sim->procs[event->kid], event );
      NextEvent(sim->fds);
      sim->tsim = event->date - sim->tmin;
      break;
    }
  }
  if (sim->discretetime ==0)
    sim->tsim += sim->hres;
  return nanosecond2second(sim->tsim);
}


/* do accumulation on data
*/
static int AccumulOneProcDataSimulator( struct Simulator* sim, int kid,
    double*      data
)
{
  /* update global information. Do not pop event */
  data[KASIM_DATA_DATE]       += nanosecond2second(sim->tsim);
  
  data[KASIM_DATA_TASK]       += (double)sim->procs[kid].taskcount;
  data[KASIM_DATA_COMPUTE]    += nanosecond2second(sim->procs[kid].tcompute);
  data[KASIM_DATA_IDLE]       += nanosecond2second(sim->procs[kid].tidle);
  data[KASIM_DATA_TASKDEEP]   += (double)sim->procs[kid].sp+1;

  sim->count_stealok[kid]     += (double)sim->procs[kid].stealcount;
  sim->efficiencies[kid]      += ProcessorEfficiencyAt( sim->procs[kid], sim->tsim+sim->tmin );

  data[KASIM_DATA_STEALOP]    += sim->count_stealok[kid];
  data[KASIM_DATA_EFFICIENCY] += sim->efficiencies[kid];

  for (int i=KASIM_DATA_PERFCTR; i<KASIM_MAX_COUNTER; ++i)
  {
    data[i] += sim->procs[kid].perfctr[0][i-KASIM_DATA_PERFCTR]+sim->procs[kid].perfctr[1][i-KASIM_DATA_PERFCTR];
  }
  return 0;
}

/* do accumulation on data
*/
static int AccumulOneStatDataSimulator( struct Simulator* sim, int kid,
    double*      data
)
{
  /* update global information. Do not pop event */
  double d = nanosecond2second_diff(sim->procs[kid].finaldate, sim->procs[kid].firstdate);
  if (data[KASIM_DATA_DURATION] < d)
    data[KASIM_DATA_DURATION] =  d;

  data[KASIM_DATA_TASK]    += (double)sim->procs[kid].taskcount;
  data[KASIM_DATA_COMPUTE] += nanosecond2second(sim->procs[kid].tcompute);
  data[KASIM_DATA_IDLE]    += nanosecond2second(sim->procs[kid].tidle);
  data[KASIM_DATA_STEALOP] += (double)sim->procs[kid].stealcount;

  double tmin = nanosecond2second(sim->procs[kid].tmin);
  double tmax = nanosecond2second(sim->procs[kid].tmax);
  if (tmin < data[KASIM_DATA_TASKMIN])
    data[KASIM_DATA_TASKMIN] = tmin;
  if (tmax > data[KASIM_DATA_TASKMAX])
    data[KASIM_DATA_TASKMAX] = tmax;
  data[KASIM_DATA_TASKAVRG]  += nanosecond2second(sim->procs[kid].tcompute);
  return 0;
}

/*
*/
extern "C"
int OneProcDataSimulator( struct Simulator* sim, int procid,
    double*      data
)
{
  for (int i=0; i<KASIM_MAX_COUNTER; ++i)
    data[i] = 0;
  return AccumulOneProcDataSimulator( sim, procid, data );
}


/*
*/
extern "C"
int ProcDataSimulator( struct Simulator* sim,
    double*      data
)
{
  for (int i=0; i<KASIM_MAX_COUNTER; ++i)
    data[i] = 0;
  
  int count = GetProcessorCount(sim->fds);
  for (int i= 0; i<count; ++i)
  {
    AccumulOneProcDataSimulator( sim, i, data );
  }
  
  /* for date, stack deep and efficiency, compute average */
  data[KASIM_DATA_DATE] /= (double)count;
  data[KASIM_DATA_TASKDEEP] /= (double)count;
  data[KASIM_DATA_EFFICIENCY] /= (double)count;
  return 0;
}

/**/
extern "C"
int OneStatDataSimulator( struct Simulator* sim, int procid,
    double*      data
)
{
  for (int i=0; i<KASIM_MAX_STATCOUNTER; ++i)
    data[i] = 0;
  data[KASIM_DATA_TASKMIN] = DBL_MAX;
  AccumulOneStatDataSimulator( sim, procid, data );
  data[KASIM_DATA_TASKAVRG]  /= data[KASIM_DATA_TASK];
  return 0;
}

/**/
extern "C"
int StatDataSimulator( struct Simulator* sim,
    double*      data
)
{
  for (int i=0; i<KASIM_MAX_STATCOUNTER; ++i)
    data[i] = 0;
  data[KASIM_DATA_TASKMIN] = DBL_MAX;
  
  int count = GetProcessorCount(sim->fds);
  for (int i= 0; i<count; ++i)
  {
    AccumulOneStatDataSimulator( sim, i, data );
  }
  
  /* for date, task time, compute average */
  data[KASIM_DATA_DATE] /= (double)count;
  data[KASIM_DATA_TASKAVRG]  /= data[KASIM_DATA_TASK];
  return 0;
}


/* Do one step of simulation and return the synthetized information
   (date, count, efficiencies[i]) for each processor.
   On return the next simulation date is date+hres (contiguous simulation mode)
   or date of the next event.
*/
extern "C"
int OneStepSimulator( Simulator*       sim,
                      double*          date,
                      int*             count, 
                      const double**   ef,
                      const uint64_t** cs
)
{
  *date = NextStepSimulator(sim);
  if (*date >=0.0)
  {
    double data[KASIM_MAX_COUNTER];
    for (int i= 0; i<KASIM_MAX_COUNTER; ++i)
      data[i] = 0.0;

    int nproc = GetProcessorCount(sim->fds);
    for (int i= 0; i<nproc; ++i)
      AccumulOneProcDataSimulator( sim, i, data );
    
    *count = nproc;
    *ef = sim->efficiencies;
    *cs = sim->count_stealok;
    return  1;
  }
  return 0;
}


/* Read and call callback on each event, ordered by date
*/
extern "C"
int CloseSimulator(Simulator* sim )
{
  if (sim ==0) return 0;
  free(sim->efficiencies);
  free(sim->count_stealok);
  free(sim->count_task);
  
  int i, count = sim->procs.size();
  for (i=0;i<count; ++i)
  {
    ProcessorState& p = sim->procs[i];
    free( p.stacktask);
    p.stackdeepth = 0;
  }

  delete sim;
  return 0;
}


/*
*/
static void processor_simulate_event(
  ProcessorState& p,
  const kaapi_event_t* event
)
{
//printf("evt:%i -> %s\n", event->evtno, kaapi_event_name[event->evtno]);
  switch (event->evtno) 
  {
    case KAAPI_EVT_KPROC_START:
      p.active   = 1;
      p.firstdate = event->date;
      p.lastdate = event->date;
      p.tidle    = 0;
      p.tactive  = 0;
      break;

    case KAAPI_EVT_KPROC_STOP:
      assert( event->date >= p.lastdate);
      p.tactive += (event->date - p.lastdate);
      assert( p.tactive <= event->date);
      p.lastdate  = event->date;
      p.finaldate = event->date;
      p.active   = 0;
      break;

    /* */
    case KAAPI_EVT_FRAME_TL_BEG:
    break;

    case KAAPI_EVT_FRAME_TL_END:
    break;

    /* unroll graph for static schedule */
    case KAAPI_EVT_STATIC_BEG:
    break;

    case KAAPI_EVT_STATIC_END:
    break;

    case KAAPI_EVT_TASKMAIN_BEG:
    case KAAPI_EVT_TASK_BEG:
    case KAAPI_EVT_STATIC_TASK_BEG:
    {
      ++p.taskcount;
      p.sp++;
      if (p.sp >= p.stackdeepth)
      {
        int newsize = p.stackdeepth *2;
        p.stacktask = (uint64_t*)realloc(p.stacktask, newsize*sizeof(uint64_t));
        if (p.stacktask ==0)
        {
          fprintf(stderr, "*** Error: cannot allocate memory\n");
          fflush(stderr);
          kaapi_abort();
        }
      }
      p.stacktask[p.sp] = event->date;
    } break;

    case KAAPI_EVT_TASKMAIN_END:
    case KAAPI_EVT_TASK_END:
    case KAAPI_EVT_STATIC_TASK_END:
    {
      if (p.sp <0) 
      {
        fprintf(stderr, "*** Invalid event End of task\n");
        fflush(stderr);
        p.sp = -1;
      }
      uint64_t delay = event->date-p.stacktask[p.sp];
      if (event->evtno == KAAPI_EVT_TASKMAIN_END)
      {
       p.tmain = delay;
      }
      else {
        if (delay < p.tmin) p.tmin = delay;
        if (delay > p.tmax) p.tmax = delay;
        if (p.sim->fd_taskdump !=-1)
        {
          char tmp[128];
          sprintf(tmp,"%10.8e %10.8e\n",nanosecond2second_diff(event->date, p.firstdate), nanosecond2second(delay));
          write(p.sim->fd_taskdump,tmp, strlen(tmp));
        }
      }
      p.tsum += delay;
      p.tcompute += delay;
      --p.sp;
    } break;

    /* idle = steal state */
    case KAAPI_EVT_SCHED_SUSPEND_BEG:
    case KAAPI_EVT_SCHED_IDLE_BEG:
      assert( p.active == 1);
      assert( event->date >= p.lastdate);
      p.tactive += (event->date - p.lastdate);
      assert( p.tactive <= event->date);
      p.lastdate = event->date;
      --p.active;
    break;

    case KAAPI_EVT_SCHED_SUSPEND_END:
    case KAAPI_EVT_SCHED_IDLE_END:
      assert( p.active == 0);
      assert( event->date >= p.lastdate);
      p.tidle   += (event->date - p.lastdate);
      assert( p.tidle <= event->date);
      p.lastdate = event->date;
      ++p.active;
    break;


    case KAAPI_EVT_SCHED_SUSPEND_POST:
    break;

    /* thread is always active when recording this event */
    case KAAPI_EVT_SCHED_SUSPWAIT_BEG:
    break;

    case KAAPI_EVT_SCHED_SUSPWAIT_END:
    break;

    case KAAPI_EVT_FOREACH_STEAL:
    break;

    case KAAPI_EVT_FOREACH_BEG:
    break;

    case KAAPI_EVT_FOREACH_END:
    break;

    /* processing request */
    case KAAPI_EVT_REQUESTS_BEG:
    break;

    case KAAPI_EVT_REQUESTS_END:
    break;

    /* emit steal */
    case KAAPI_EVT_STEAL_OP:
    break;

    /* emit reply */
    case KAAPI_EVT_SEND_REPLY:
    break;

    /* recv reply */
    case KAAPI_EVT_RECV_REPLY:
      if (event->d2.i !=0)
        ++p.stealcount;
    break;

    case KAAPI_EVT_PERFCOUNTER:
      p.perfctr[KAAPI_PERF_USER_STATE][event->d0.i] = event->d1.u;
      p.perfctr[KAAPI_PERF_SCHEDULE_STATE][event->d0.i] = event->d2.u;
    break;
    
    default:
      printf("***Unkown event number: %i\n", event->evtno);
      break;
  }
}
