/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** Thierry Gautier, thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef KAAPI_TRACE_SIMULATOR_H
#define KAAPI_TRACE_SIMULATOR_H

#include "kaapi_trace_reader.h"

#if defined(__cplusplus)
extern "C" {
#endif

/* Mask for data generated by the simulator */
#define KASIM_MASK_TASK        0x1    /* report task count over time */
#define KASIM_MASK_COMPUTE     0x2    /* report cumul of activity since the startup */
#define KASIM_MASK_IDLE        0x4    /* report cumul of idle time since the startup */
#define KASIM_MASK_TASKDEEP    0x8    /* report stack deep over the time */
#define KASIM_MASK_STEALOP     0x10   /* report steal ok count over the time */
#define KASIM_MASK_EFFICIENCY  0x20   /* report efficiency over the time */
#define KASIM_MASK_TASKDUMP    0x40   /* dump task times */
#define KASIM_MASK_PERFCTR     0x80   /* dump perf counters */


/* Set of files 
*/
struct Simulator;

/* Open a set of event files
*/
extern struct Simulator* OpenSimulator( struct FileSet* fds, double hres, int discrete, int dumptasktimes );

/* Do one step of simulation and return the synthetized information
   (date, count, efficiencies[i]) for each processor.
   Return 0 if no more event to simulate.
   *date = *count = 0 iff event set is empty.
*/
extern int OneStepSimulator( struct Simulator* sim,
                             double* date,
                             int* count, 
                             const double** efficiencies,
                             const uint64_t** count_stealok
                           );

/** Advance to the next step
    Return the current date of simulation.
    Return 0.0 when no more simulation step is available
*/
extern double NextStepSimulator( struct Simulator* sim );

#define KASIM_MAX_COUNTER       (7 + KAAPI_PERF_ID_MAX)
#define KASIM_MAX_STATCOUNTER   8

/* Format of data */
#define KASIM_DATA_DATE        0    /* report current date */
#define KASIM_DATA_TASK        1    /* report task count over time */
#define KASIM_DATA_COMPUTE     2    /* report cumul of activity since the startup */
#define KASIM_DATA_IDLE        3    /* report cumul of idle time since the startup */
#define KASIM_DATA_STEALOP     4   /* report steal ok count over the time */
#define KASIM_DATA_TASKDEEP    5    /* report stack deep over the time */
#define KASIM_DATA_EFFICIENCY  6   /* report efficiency over the time */
#define KASIM_DATA_PERFCTR     7   /* perfctr offset */

/* Format of data */
#define KASIM_DATA_DURATION    0    /* at the end: duration */
#define KASIM_DATA_TASKMIN     5    /* report task count over time */
#define KASIM_DATA_TASKMAX     6    /* report cumul of activity since the startup */
#define KASIM_DATA_TASKAVRG    7    /* report cumul of activity since the startup */


/* Cumulate data to all processors 
   data must have size KASIM_MAX_COUNTER  and the format is:
   data[0] = date
   data[1] = #task
   data[2] = tcompute
   data[3] = tidle
   data[4] = #stealop
   data[5] = #stack deep
   data[6] = efficiency
   data[7]..data[KASIM_MAX_COUNTER] the performance counters.
   
   If the method is called after NextStepSimulator() returns 0.0,
   then the data contains the last value which is the cumul over all the execution.
*/
extern int ProcDataSimulator( struct Simulator* sim,
    double*      data
);

/* Return information for one kprocessor.
   Same format of data as in function ProcDataSimulator
*/
extern int OneProcDataSimulator( struct Simulator* sim, int procid,
    double*      data
);

/* Cumulate data to all processors
   data must have size KASIM_MAX_STATCOUNTER and the format is:
   data[0] = duration
   data[1] = #task
   data[2] = tcompute
   data[3] = tidle
   data[4] = #stealop
   data[5] = tmin
   data[6] = tmax

   
   If the method is called after NextStepSimulator() returns 0.0,
   then the data contains the last value which is the cumul over all the execution.
*/
extern int StatDataSimulator( struct Simulator* sim,
    double*      data
);

/* Return information for one kprocessor.
   Same format of data as in function ProcDataSimulator
*/
extern int OneStatDataSimulator( struct Simulator* sim, int procid,
    double*      data
);


/* Read and call callback on each event, ordered by date
*/
extern int CloseSimulator(struct Simulator* sim );

/** Return nanosecond to second 
*/
static inline double nanosecond2second( uint64_t ns)
{ return 1.0e-9*(double)ns; }

/** Return nanosecond to second 
*/
static inline double nanosecond2second_diff( uint64_t t1, uint64_t t0)
{ return 1.0e-9*(double)(t1-t0); }

#if defined(__cplusplus)
}
#endif

#endif