#ifndef _KAAPI_SCHED_H_
#define _KAAPI_SCHED_H_

#include "kaapi_trace.h"

typedef struct kaapi_sched_t {
  void                (*init)           (void);
  void                (*push)           (kaapi_tasklist_t*	const, kaapi_taskdescr_t* const);
  int                 (*pop)            ( kaapi_readytasklist_t*, kaapi_taskdescr_t** );
  uint32_t            (*push_activated) (kaapi_tasklist_t*	const, kaapi_taskdescr_t* const);
  kaapi_taskdescr_t*  (*steal)          (kaapi_processor_t* const, kaapi_taskdescr_t* const);
  void                (*prologue)       (kaapi_taskdescr_t* const);
  void                (*epilogue)       (kaapi_taskdescr_t* const);
  void                (*finalize)       (void);
} kaapi_sched_t;

extern kaapi_sched_t kaapi_scheduler;

#if defined(KAAPI_USE_PERFCOUNTER)
extern kaapi_atomic_t kaapi_sched_perfcounter_enabled;

static inline int kaapi_sched_trace_is_enabled()
{
  return (KAAPI_ATOMIC_READ(&kaapi_sched_perfcounter_enabled) != 0);
}
#endif

extern void kaapi_sched_init(void);

extern void kaapi_sched_finalize(void);

static inline uint32_t kaapi_sched_push_activated( kaapi_tasklist_t*	const tasklist,
                                      kaapi_taskdescr_t* const td )
{
  if(kaapi_scheduler.push_activated == 0)
    return 0;
  
#if defined(KAAPI_USE_PERFCOUNTER)
  if(kaapi_sched_trace_is_enabled()){
    kaapi_perf_counter_t date0 = 0;
    kaapi_perf_counter_t date1 = 0;
    uint32_t cnt= 0;
    date0 = kaapi_get_elapsedns();
    cnt = kaapi_scheduler.push_activated(tasklist, td);
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_REG_SYS(kaapi_get_current_processor(), KAAPI_PERF_ID_SCHED_PUSH) += date1-date0;
    return cnt;
  } else
#endif
   return kaapi_scheduler.push_activated(tasklist, td);
}

static inline void kaapi_sched_push( kaapi_tasklist_t*	const tasklist,
                                kaapi_taskdescr_t* const td )
{
  if(kaapi_scheduler.push == 0)
    return;
  
#if defined(KAAPI_USE_PERFCOUNTER)
  if(kaapi_sched_trace_is_enabled()){
    kaapi_perf_counter_t date0 = 0;
    kaapi_perf_counter_t date1 = 0;
    date0 = kaapi_get_elapsedns();
    kaapi_scheduler.push(tasklist, td);
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_REG_SYS(kaapi_get_current_processor(), KAAPI_PERF_ID_SCHED_PUSH) += date1-date0;
  } else
#endif  
  kaapi_scheduler.push(tasklist, td);
}

static inline int kaapi_sched_pop( kaapi_readytasklist_t* rtl, kaapi_taskdescr_t** td )
{
  if(kaapi_scheduler.pop == 0)
    return 1; /* equivalent to no tasks */
  
  /* ensure readylist has any/some task */
  if (kaapi_readytasklist_isempty(rtl))
    return 1;
  
#if defined(KAAPI_USE_PERFCOUNTER)
  if(kaapi_sched_trace_is_enabled()){
    kaapi_perf_counter_t date0 = 0;
    kaapi_perf_counter_t date1 = 0;
    int ret = 0;
    date0 = kaapi_get_elapsedns();
    ret = kaapi_scheduler.pop(rtl, td);
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_REG_SYS(kaapi_get_current_processor(), KAAPI_PERF_ID_SCHED_POP) += date1-date0;
    return ret;
  } else
#endif
  return kaapi_scheduler.pop(rtl, td);
}

static inline kaapi_taskdescr_t* kaapi_sched_steal(kaapi_processor_t* const thief, kaapi_taskdescr_t* const td)
{
  if(kaapi_scheduler.steal == 0)
    return 0;
  
#if defined(KAAPI_USE_PERFCOUNTER)
  if(kaapi_sched_trace_is_enabled()){
    kaapi_perf_counter_t date0 = 0;
    kaapi_perf_counter_t date1 = 0;
    kaapi_taskdescr_t* _td = 0;
    date0 = kaapi_get_elapsedns();
     _td = kaapi_scheduler.steal(thief, td);
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_REG_SYS(kaapi_get_current_processor(), KAAPI_PERF_ID_SCHED_STEAL) += date1-date0;
    return _td;
  } else
#endif  
  return kaapi_scheduler.steal(thief, td);
}

static inline void kaapi_sched_task_epilogue(kaapi_taskdescr_t* const td)
{
  if(kaapi_scheduler.epilogue)
    kaapi_scheduler.epilogue(td);
}

static inline void kaapi_sched_task_prologue(kaapi_taskdescr_t* const td)
{
  if(kaapi_scheduler.prologue)
    kaapi_scheduler.prologue(td);
}

static inline int kaapi_sched_push_by_site(kaapi_taskdescr_t* const td)
{
  /* always take the first bit ==1 */
  int kid = kaapi_taskdescr_get_site(td)-1;
  kaapi_assert_debug( (kid >=0) && (kid < (int32_t)kaapi_count_kprocessors) );
  kaapi_processor_t* kproc_remote = kaapi_all_kprocessors[kid];
  /* forbid steal of this task */
  kaapi_task_set_unstealable(td->task);
  return kaapi_readylist_remote_push( kproc_remote->rtl, td );
}

/** Used to steal a tasklist in the frame
*/
extern void kaapi_sched_stealtasklist( 
                           kaapi_thread_context_t*       thread, 
                           kaapi_tasklist_t*             tasklist, 
                           kaapi_listrequest_t*          lrequests, 
                           kaapi_listrequest_iterator_t* lrrange
);

/** Used to steal a tasklist from local readylist of victim
*/
extern void kaapi_sched_stealreadytasklist( 
                           kaapi_thread_context_t*       thread, 
                           kaapi_processor_t*            proc, 
                           kaapi_listrequest_t*          lrequests, 
                           kaapi_listrequest_iterator_t* lrrange
);

extern void kaapi_sched_stealreadytasklist_priority( 
                           kaapi_thread_context_t*       thread, 
                           kaapi_processor_t*            kproc,
                           kaapi_listrequest_t*          lrequests, 
                           kaapi_listrequest_iterator_t* lrrange
);

/* try to pop a task in priority queue of the current kproc (assumed to be idle in sched_idle
   or sched_suspend).
   The task is pushed in the thread top frame and execution must be done using kaapi_thread_execframe 
   function.
*/
extern int kaapi_sched_popreadytasklist_priority(
                           kaapi_thread_context_t*       thread, 
                           kaapi_processor_t*            kproc
);


/** Start execution with a TD.
    Initialize the stack of the current thread and then, the caller shoud calls kaapi_thread_execframe_tasklist.
    Return 0 in case of success.
*/
extern int kaapi_thread_startexecwithtd( 
      kaapi_processor_t* kproc,
      kaapi_taskdescr_t* td
);

int kaapi_thread_execframe_tasklist( kaapi_thread_context_t* thread );

#endif /* _KAAPI_SCHED_H_ */
