/*
** kaapi_cuda_proc.h
** xkaapi
** 
** Created on Jul 2010
** Copyright 2010 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@imag.fr
** Joao.Lima@imagf.r / joao.lima@inf.ufrgs.br 
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#include <stdio.h>
#include <cuda_runtime_api.h>

#include "kaapi_impl.h"
#include "kaapi_cuda_proc.h"
#include "kaapi_cuda_dev.h"
#include "kaapi_cuda_ctx.h"
#include "kaapi_cuda_cublas.h"
#include "kaapi_cuda_mem.h"
#include "kaapi_cuda_mem_cache.h"

#include "kaapi_cuda_stream.h"

#if defined(KAAPI_USE_CUPTI)
#include "kaapi_cuda_trace.h"
#endif

#if 0
#  include "hwloc.h"
#  include "hwloc/cudart.h"
#endif

/* total number of CUDA devices */
uint32_t kaapi_cuda_count_kprocessors = 0;

/* number of CUDA devices per kprocessor */
static uint32_t kaapi_cuda_count_by_kprocessors[KAAPI_MAX_PROCESSOR];

/* index of kprocessors by CUDA devid */
kaapi_processor_id_t kaapi_cuda_all_kprocessors[KAAPI_CUDA_MAX_DEV];

kaapi_cuda_proc_t* kaapi_cuda_proc_all[KAAPI_MAX_PROCESSOR];

int kaapi_cuda_get_devices_by_kprocessor(kaapi_processor_id_t kid)
{
  return kaapi_cuda_count_by_kprocessors[kid];
}

void kaapi_cuda_init(void)
{
  KAAPI_ATOMIC_WRITE(&kaapi_cuda_synchronize_barrier, 0);
  memset(kaapi_cuda_all_kprocessors, 0, KAAPI_CUDA_MAX_DEV*sizeof(kaapi_processor_id_t));
  memset(kaapi_cuda_proc_all, 0, KAAPI_MAX_PROCESSOR*sizeof(kaapi_cuda_proc_t*));
  memset(kaapi_cuda_count_by_kprocessors, 0, KAAPI_MAX_PROCESSOR*sizeof(uint32_t));
  
#if defined(KAAPI_USE_PERFMODEL)
  /* Warn user about performance loss */
  if(kaapi_perfmodel_calibrate_is_enabled())
  {
    fprintf(stderr, "%s:%d:%s: *** Kaapi: WARNING performance model calibration is enabled. It may significantly affect GPU performance.\n",
            __FILE__, __LINE__, __FUNCTION__
            );
    fflush(stderr);
  }
#endif
  
#if 0
  int dev_count;
  int dev;
  int err;
  
  cudaGetDeviceCount(&dev_count);
  hwloc_topology_t topology;
  hwloc_topology_init(&topology);
  hwloc_topology_load(topology);
  for (dev = 0; dev < dev_count; dev++)
  {
    hwloc_bitmap_t set;
    set = hwloc_bitmap_alloc();
    err = hwloc_cudart_get_device_cpuset(topology, dev, set);
    if (err < 0) {
      fprintf(stdout, "failed to get cpuset for device %d\n", dev);
    } else {
      char *cpuset_string = NULL;
      hwloc_bitmap_asprintf(&cpuset_string, set);
      fprintf(stdout, "[%s] got cpuset %s for device %d\n", __FUNCTION__, cpuset_string, dev);
      free(cpuset_string);
    }
    hwloc_bitmap_free(set);
  }
  fflush(stdout);
  hwloc_topology_destroy(topology);  
#endif
}

int kaapi_cuda_proc_init(kaapi_processor_t* const kproc, const unsigned int idev, const unsigned int ndevices)
{
  int ctx;
  kaapi_cuda_proc_t* cuda_proc;

  if(kaapi_cuda_proc_all[kproc->kid] == 0)
    kaapi_cuda_proc_all[kproc->kid] = kaapi_cuda_proc_alloc();
  
  cuda_proc = kaapi_cuda_proc_all[kproc->kid];
  cuda_proc->is_initialized = 0;

  for(ctx = 0; ctx < ndevices; ctx++)
  {
    int dev = idev + ctx; /* CUDA device Id to initilize */
    cuda_proc->index = ctx;
    kaapi_cuda_ctx_init(&cuda_proc->ctx[ctx], dev);
    kaapi_cuda_all_kprocessors[dev] = kproc->kid;
    kaapi_cuda_count_by_kprocessors[kproc->kid]++;    
  }
  
  KAAPI_ATOMIC_WRITE(&cuda_proc->synchronize_flag, 0);
  cuda_proc->is_initialized = 1;
  kaapi_cuda_count_kprocessors++;
  
  return 0;
}


int kaapi_cuda_proc_cleanup(kaapi_processor_t* const kproc)
{
  if (kaapi_processor_get_cudaproc(kproc)->is_initialized == 0)
    return -1;

  kaapi_processor_get_cudaproc(kproc)->is_initialized = 0;
  free(kaapi_cuda_proc_all[kproc->kid]);
  kaapi_cuda_proc_all[kproc->kid] = 0;

  return 0;
}

int kaapi_getconcurrency_gpu(void)
{
  return kaapi_cuda_count_kprocessors;
}

cudaStream_t kaapi_cuda_kernel_stream(void)
{
  kaapi_cuda_ctx_t* const ctx = kaapi_cuda_ctx_get_current();
  return kaapi_cuda_get_cudastream(kaapi_cuda_get_kernel_fifo(ctx->kstream));
}

cudaStream_t kaapi_cuda_HtoD_stream(void)
{
  kaapi_cuda_ctx_t* const ctx = kaapi_cuda_ctx_get_current();
  return kaapi_cuda_get_cudastream(kaapi_cuda_get_input_fifo(ctx->kstream));
}

cudaStream_t kaapi_cuda_DtoH_stream(void)
{
  kaapi_cuda_ctx_t* const ctx = kaapi_cuda_ctx_get_current();
  return kaapi_cuda_get_cudastream(kaapi_cuda_get_output_fifo(ctx->kstream));
}

cudaStream_t kaapi_cuda_DtoD_stream(void)
{
  kaapi_cuda_ctx_t* const ctx = kaapi_cuda_ctx_get_current();
  return kaapi_cuda_get_cudastream(kaapi_cuda_get_output_fifo(ctx->kstream));  
}

void kaapi_cuda_proc_poll(kaapi_processor_t * const kproc)
{
  if (KAAPI_ATOMIC_READ(&kaapi_processor_get_cudaproc(kproc)->synchronize_flag) == 1) {
    kaapi_cuda_sync(kproc);
    KAAPI_ATOMIC_WRITE(&kaapi_processor_get_cudaproc(kproc)->synchronize_flag, 0);
  } else {
    kaapi_cuda_stream_poll(kproc);
  }
}

int kaapi_cuda_proc_end_isvalid(kaapi_processor_t * const kproc)
{
  const int ndevices = kaapi_cuda_get_devices_by_kprocessor(kproc->kid);
  int ctx;
  
  for(ctx = 0; ctx < ndevices; ctx++)
  {
    if(!kaapi_cuda_stream_is_empty(kaapi_processor_get_cudaproc(kproc)->ctx[ctx].kstream))
      return 0;
  }
  
  return 1;
}

int kaapi_cuda_proc_all_isvalid(void)
{
  kaapi_processor_t **pos = kaapi_all_kprocessors;
  size_t i;

  for (i = 0; i < kaapi_count_kprocessors; ++i, ++pos)
    if ((*pos)->proc_type == KAAPI_PROC_TYPE_CUDA) {
      if (!kaapi_cuda_proc_end_isvalid(*pos))
          return 0;
    }

  return 1;
}

void kaapi_cuda_proc_destroy(kaapi_processor_t* const kproc)
{
#if defined(KAAPI_USE_CUPTI)
  kaapi_atomic_destroylock(&kaapi_processor_get_cudaproc(kproc)->ctx.lock);
#endif
#if 0
  kaapi_cuda_cublas_finalize(&kaapi_processor_get_cudaproc(kproc));
  kaapi_cuda_stream_destroy(kaapi_processor_get_cudaproc(kproc)->kstream);
  kaapi_cuda_mem_destroy(&kaapi_processor_get_cudaproc(kproc));
  kaapi_cuda_dev_close(&kaapi_processor_get_cudaproc(kproc));
#endif
}

int kaapi_cuda_self_device(void)
{
  kaapi_cuda_ctx_t* const ctx = kaapi_cuda_ctx_get_current();
  return ctx->dev;
}

#if defined(KAAPI_DEBUG)
int kaapi_cuda_check_current_device(void)
{
  int cuda_dev;
  int kaapi_dev;
  
  const cudaError_t res = cudaGetDevice(&cuda_dev);
  if (res != cudaSuccess) {
    fprintf(stdout, "%s: cudaGetDevice ERROR %d\n", __FUNCTION__, res);
    fflush(stdout);
    return (int)res;
  }
  kaapi_dev = kaapi_cuda_ctx_get_current_device();
  if( kaapi_dev != cuda_dev)
  {
    fprintf(stderr, "*** (%s) Kaapi runtime ERROR: current CUDA device is different from Kaapi context (Kaapi device is '%d' while CUDA returned '%d'). \n",
            __FUNCTION__, kaapi_dev, cuda_dev);
    fflush(stderr);
    return 1;
  }
  
  return 0;
}

#endif
