/*
 ** xkaapi
 ** 
 ** Copyright 2-110 INRIA.
 **
 ** Contributors :
 **
 ** fabien.lementec@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "kaapi_impl.h"

#if defined(KAAPI_USE_NUMA)
#include <sys/mman.h>
#include <unistd.h>
#include <numaif.h>
#include <numa.h>
#endif

/* Convert public numanode representation (ranging from 0 to #NUMAnode-1)
   to OS numa node identifier.
   kaapi_ld2osid[ kaapi identifier ] = OS id of the corresponding numa node or -1 if unused
   kaapi_osid2ld[ numa node ] = the corresponding kaapi identifier or -1 if unused
*/
static int kaapi_ld2osid[64] = {
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

static int kaapi_osid2ld[64] = {
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};



/*
*/
kaapi_ldid_t kaapi_ldset_getpage_id(const void* addr)
{
#if defined(KAAPI_USE_NUMA)
  int mode;
  const int err = get_mempolicy(&mode, 0, 0, (void* )addr, MPOL_F_NODE | MPOL_F_ADDR);

  if (err)
  {
    return (kaapi_ldid_t)-1;
  }

//printf("Node for@:%p, on os_idx:%i, ldid:%i\n", addr, mode, kaapi_osid2ld[mode] );

  /* convert to internal kaapi identifier */
  return kaapi_osid2ld[mode];
#else
  return 0;
#endif
}


/*
*/
kaapi_ldset_t kaapi_ldset_getpage_ids(const void* addr, size_t size)
{
  kaapi_ldid_t retval = kaapi_ldset_getpage_id(addr);
  if (retval !=(kaapi_ldid_t)-1)
    return 1UL << retval;
  return 0UL;
}

/*
*/
int kaapi_ldset_bind_bloc1dcyclic(kaapi_ldset_t ldset, const void* addr, size_t size, size_t blocsize)
{
#if defined(KAAPI_USE_NUMA)
  const int mode = MPOL_BIND;
  const unsigned int flags = MPOL_MF_STRICT | MPOL_MF_MOVE;
  const char* base = (const char*)addr;
  kaapi_bitmap_value64_t numaset;
  size_t pagesize = getpagesize();
  const unsigned long maxnode = KAAPI_MAX_PROCESSOR;
  unsigned long nodemask
      [(KAAPI_MAX_PROCESSOR + (8 * sizeof(unsigned long) -1))/ (8 * sizeof(unsigned long))];

  if ((ldset ==0) || ((blocsize & (pagesize-1)) !=0) /* should be divisible by getpagesize() */)
  {
    errno = EINVAL;
    return -1;
  }
  if (((uintptr_t)addr & (pagesize-1)) !=0) /* should aligned on page boundary */
  {
    errno = EFAULT;
    return -1;
  }
  
  kaapi_bitmap_value_init_64(&numaset, &ldset);
  int node = 0;
  while (size >0)
  {
    int ldid = kaapi_bitmap_value_first1_and_zero_64( &numaset );
    if (ldid ==0)
    {
      kaapi_bitmap_value_init_64(&numaset, &ldset);
      continue;
    }
    --ldid;
    node = kaapi_ld2osid[ldid];

    memset(nodemask, 0, sizeof(nodemask));
    nodemask[node / (8 * sizeof(unsigned long))] |= 1UL << (node % (8 * sizeof(unsigned long)));

#if 0//defined(KAAPI_DEBUG)
  /* avoid access */
  if (mprotect( (void*)base, blocsize, PROT_NONE))
  {
    printf("[kaapi_ldset_bind_bloc1dcyclic] cannot set protection: %s\n", strerror(errno) );
    return -1;
  }
#endif

    if (mbind((void*)base, blocsize, mode, nodemask, maxnode, flags))
      return -1;
    //printf("Bind @:%p, blocsize:%li, os_idx:%i, ldid:%i\n", (void*)base, (long)blocsize, (int)node, (int)ldid );

#if 0//defined(KAAPI_DEBUG)
  /* avoid access */
  if (mprotect( (void*)base, blocsize, PROT_READ|PROT_WRITE))
  {
    printf("[kaapi_ldset_bind_bloc1dcyclic] cannot set protection: %s\n", strerror(errno) );
    return -1;
  }
#endif

    base += blocsize;
    if (size > blocsize)
      size -= blocsize;
    else
      size = 0;
  }
  /* add mbind for an extra page */
  if (mbind((void*)base, pagesize, mode, nodemask, maxnode, flags))
  {
      printf("*** Error in binding past the last page\n");
      return -1;
  }
#endif

  return 0;
}


kaapi_ldset_t kaapi_all_lds(kaapi_hws_levelid_t level)
{
  return (1U << kaapi_queue_count_perlevel(level))-1;
}


/*
*/
unsigned int kaapi_ldset_count( kaapi_hws_levelid_t level )
{
  return kaapi_queue_count_perlevel(level);
}


/*
*/
kaapi_ldid_t kaapi_ldset_get_current(kaapi_hws_levelid_t level)
{
  if ((level < KAAPI_HWS_LEVELID_L0) || (level >KAAPI_HWS_LEVELID_MACHINE))
    return (kaapi_ldid_t)-1;
  kaapi_processor_t* kproc = kaapi_get_current_processor();
  if (kproc->highpriority_queue[level] ==0)
    return (kaapi_ldid_t)-1;
  return kproc->highpriority_queue[level]->lid;
}


/*
*/
kaapi_ldid_t kaapi_ldset_get( kaapi_hws_levelid_t level, int i )
{
  if ((level < KAAPI_HWS_LEVELID_L0) || (level >KAAPI_HWS_LEVELID_MACHINE))
    return (kaapi_ldid_t)-1;
  return kaapi_queue_getperlevel(level,i)->lid;
}

int kaapi_ldset_get_cpuset( kaapi_hws_levelid_t level, int i, kaapi_cpuset_t* cpuset )
{
  if ((level < KAAPI_HWS_LEVELID_L0) || (level >KAAPI_HWS_LEVELID_MACHINE))
    return EINVAL;
  *cpuset = kaapi_queue_getperlevel(level,i)->cpuset;
  return 0;
}


int kaapi_ldset_count_cpu(kaapi_hws_levelid_t level, int i)
{
  if ((level < KAAPI_HWS_LEVELID_L0) || (level >KAAPI_HWS_LEVELID_MACHINE))
  {
    errno = EINVAL;
    return -1;
  }
  return kaapi_cpuset_count( &kaapi_queue_getperlevel(level,i)->cpuset );
}

/*
*/
void* kaapi_ldset_alloc(size_t size)
{
  void* retval;
#if defined(KAAPI_USE_NUMA)
  size_t pagesize, size_mmap;
  
  pagesize = getpagesize();
  size_mmap = (( size + pagesize -1 ) / pagesize) * pagesize + pagesize;
  retval = mmap( 0, size_mmap, PROT_READ|PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, (off_t)0 );
  if (retval == (void*)-1)
    return 0;
//printf("Allocate @:%p, sizemmap:%lu, requested size:%lu\n", retval, (long)size_mmap, (long)size ); 
#else
 retval = malloc( size );
#endif
  return retval;
}


/*
*/
int kaapi_ldset_free(void* ptr, size_t size)
{
#if defined(KAAPI_USE_NUMA)
  size_t pagesize, size_mmap;
  pagesize = getpagesize();

  if (((uintptr_t)ptr & (pagesize-1)) !=0) /* should aligned on page boundary */
  {
    errno = EFAULT;
    return -1;
  }
  size_mmap = (( size + pagesize -1 ) / pagesize) * pagesize + pagesize;
  munmap( ptr, size_mmap );
#else
  free(ptr);
#endif
  return 0;
}


/*
*/
void* kaapi_ldset_allocate_bloc1dcyclic(kaapi_ldset_t ldset, size_t size, size_t blocsize)
{
  int err;
  void* retval;

  retval = kaapi_ldset_alloc(size);
  if (retval == 0)
    return 0;

  err = kaapi_ldset_bind_bloc1dcyclic( ldset, retval, size, blocsize );
  if (err !=0)
  {
    kaapi_ldset_free( retval, size );
    return 0;
  }
  return retval;
}

/*
*/
int kaapi_ldset_bind(const void* addr, size_t size, int ld)
{
#if defined(KAAPI_USE_NUMA)
  const int mode = MPOL_BIND;
  const unsigned int flags = MPOL_MF_STRICT | MPOL_MF_MOVE;
  const unsigned long maxnode = KAAPI_MAX_PROCESSOR;
  
  if ((size & 0xfff) !=0) /* should be divisible by 4096 */
  {
    errno = EINVAL;
    return -1;
  }
  if (((uintptr_t)addr & 0xfff) !=0) /* should aligned on page boundary */
  {
    errno = EFAULT;
    return -1;
  }
  
  int node = kaapi_ld2osid[ld];
  unsigned long nodemask
    [(KAAPI_MAX_PROCESSOR + (8 * sizeof(unsigned long) -1))/ (8 * sizeof(unsigned long))];

  memset(nodemask, 0, sizeof(nodemask));

  nodemask[node / (8 * sizeof(unsigned long))] |= 1UL << (node % (8 * sizeof(unsigned long)));

  if (mbind((void*)addr, size, mode, nodemask, maxnode, flags))
    return -1;
#endif
  return 0;
}



/* Initialize mapping ldid <-> os_index that follows the order of bits (from low order
   to high order).
*/
int kaapi_ldset_initialize(void)
{
  unsigned int i;
  unsigned int j;
  unsigned int sizebits;
  kaapi_ldset_t ldset_all = 0;
#if defined(KAAPI_USE_NUMA)
  struct bitmask* allowed_numanode;
  numa_set_bind_policy(1);
  numa_set_strict(1);
#else
  unsigned int allowed_numanode;
#endif
  
  /* mapping : Kaapi numa node id to OS identifier */
#if defined(KAAPI_USE_NUMA)
  allowed_numanode = numa_get_mems_allowed();
  sizebits = 8*numa_bitmask_nbytes(allowed_numanode);
#else
  allowed_numanode = 1;
  sizebits = 1;
#endif
  for (j=0, i=0; i< sizebits; ++i)
  {
#if defined(KAAPI_USE_NUMA)
    if (numa_bitmask_isbitset(allowed_numanode, i))
#else
    if ((allowed_numanode & (1U << i)) !=0)
#endif
    {
      ldset_all |= (1<<j);
      kaapi_ld2osid[j] = i;
      kaapi_osid2ld[i] = j;
      ++j;
    }
  }

  return 0;
}


/* 
*/
int kaapi_ldset_swapids( kaapi_ldid_t ldid, int os_index )
{
  if (ldid >63) 
    return EINVAL;
  if (kaapi_ld2osid[ldid] == os_index)
    return 0;
  kaapi_ld2osid[ldid] = os_index;
  kaapi_osid2ld[os_index] = ldid;
  return 0;
}


/*
*/
int kaapi_ldset_finalize(void)
{
  return 0;
}


/* internal function, but extern
*/
void kaapi_dump_numa_info()
{
#if 1
  int j;
  int size = 8;
  printf("Mapping ldid->os_index:\n");
  for (j=0; j<size; ++j)
    printf("ldid:%i  -> os_idx:%i\n", j, kaapi_ld2osid[j]);

  printf("Mapping os_index->ldid:\n");
  for (j=0; j<size; ++j)
    printf("os_idx:%i  -> ldid:%i\n", j, kaapi_osid2ld[j]);
#endif
}

