/*
** xkaapi
** 
** Copyright 2013 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#include "kaapi_impl.h"

/* Constant to define the corresponding mapps queue_id -> queue.
   This size should be enough to keep track :
     - number of NUMA nodes
     - number of SOCKET (such as AMD Istanbul processors with 2 NUMA nodes per socket)
     - number of L3 cache
*/
#define KAAPI_MAX_QUEUE 512

/* Atomic counter to name queue */
static kaapi_atomic_t count_id_queue = {0};
static kaapi_atomic_t count_idperlevel_queue[KAAPI_HWS_LEVELID_MAX] = 
{
  {0},
  {0},
  {0},
  {0},
  {0}
};

/* mapping queue_id -> queue */
static kaapi_queue_t* kaapi_all_queues[KAAPI_MAX_QUEUE];
static kaapi_queue_t* kaapi_allperlevel_queues[KAAPI_HWS_LEVELID_MAX][KAAPI_MAX_QUEUE];

kaapi_queue_t* kaapi_queue_get( int id )
{
  if ((id<0) || (id > KAAPI_ATOMIC_READ(&count_id_queue)))
    return 0;
  return kaapi_all_queues[id];
}

size_t kaapi_queue_count_perlevel( kaapi_hws_levelid_t level)
{
  if ((level <0) || (level >=KAAPI_HWS_LEVELID_MAX))
    return (size_t)-1;
  return (size_t)KAAPI_ATOMIC_READ(&count_idperlevel_queue[level]);
}

kaapi_queue_t* kaapi_queue_getperlevel( kaapi_hws_levelid_t level, int id )
{
  if ((level <0) || (level >=KAAPI_HWS_LEVELID_MAX))
    return 0;
  if ((id<0) || (id > KAAPI_ATOMIC_READ(&count_idperlevel_queue[level])))
    return 0;
  return kaapi_allperlevel_queues[level][id];
}

int kaapi_queue_attach( kaapi_processor_id_t kid, kaapi_queue_t* queue )
{
  kaapi_cpuset_set( &queue->cpuset, kid );
  return 0;
}

int kaapi_queue_init( kaapi_queue_t* q, kaapi_hws_levelid_t level, int os_idx )
{ 
  if ((level <0) || (level >=KAAPI_HWS_LEVELID_MAX))
    return EINVAL;
  q->initialbloc.next   = 0;
  q->initialbloc.prev   = 0;
  q->initialbloc.top    = 0;
  q->initialbloc.bottom = 0;
  q->os_idx             = os_idx;
  kaapi_cpuset_clear( &q->cpuset );
  q->id                 = KAAPI_ATOMIC_INCR(&count_id_queue)-1;
  q->lid                = KAAPI_ATOMIC_INCR(&count_idperlevel_queue[level])-1;
  kaapi_assert_debug( 0 <= q->id );
  if ( q->id >= KAAPI_MAX_QUEUE)
    return ENOMEM;
  kaapi_all_queues[q->id] = q;
  kaapi_allperlevel_queues[level][q->lid] = q;
  q->head = q->tail = &q->initialbloc; q->freelist = 0;
  if (level == KAAPI_HWS_LEVELID_NUMA)
    kaapi_ldset_swapids( q->lid, os_idx );
  return kaapi_sched_initlock(&q->lock);
}


/* */
int kaapi_queue_destroy( kaapi_queue_t* q )
{ 
  kaapi_assert_debug(q->head == 0);
  kaapi_taskbloc_t *bloc = q->freelist;
  while (bloc !=0)
  {
    kaapi_taskbloc_t* next = bloc->next;
    if (next != &q->initialbloc)
      free(bloc);
    bloc = next;
  }
  kaapi_all_queues[q->id] = 0;
  return kaapi_sched_destroylock(&q->lock);
}
