
#include <stdio.h>
#include <math.h>
#include <unistd.h>

#if defined (_WIN32) || defined(__APPLE__)
#include <sys/sysctl.h>
#endif

#include "kaapi_impl.h"

#if defined(KAAPI_USE_CUDA)

#include <cuda_runtime_api.h>
#include "machine/cuda/kaapi_cuda_impl.h"

#define KAAPI_PERFMODEL_COMM_FILE    "kaapi_perfmodel_comm_bandwidth_dump"
#define KAAPI_PERFMODEL_MAX_PARAMS    6

#define SIZE  (32*1024*1024*sizeof(char)) /* MB */
#define NITER (90)

/* All performance results are in microseconds (us) */

typedef struct kaapi_bandwidth_entry_t {
  double latency;
  double bandwidth;
} kaapi_bandwidth_entry_t;

typedef struct kaapi_pair_entry_t {
  kaapi_bandwidth_entry_t dev[KAAPI_MAX_PROCESSOR][KAAPI_MAX_PROCESSOR];
} kaapi_pair_entry_t;

static kaapi_pair_entry_t kaapi_machine_model[KAAPI_PROC_TYPE_MAX][KAAPI_PROC_TYPE_MAX];

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

static inline char* kaapi_get_arch_string(int arch)
{
  switch (arch) {
    case KAAPI_PROC_TYPE_CPU:
      return "CPU";
      break;
      
    case KAAPI_PROC_TYPE_CUDA:
      return "CUDA";
      
    case KAAPI_PROC_TYPE_MPSOC:
      return "MPSOC";
      
    default:
      break;
  }
  
  return "none";
}

static inline int kaapi_get_string_arch(char* const str)
{
  if(strcmp(str, "CPU") == 0)
    return KAAPI_PROC_TYPE_CPU;
  
  if(strcmp(str, "CUDA") == 0)
    return KAAPI_PROC_TYPE_CUDA;
  
  if(strcmp(str, "MPSOC") == 0)
    return KAAPI_PROC_TYPE_MPSOC;
  
  return KAAPI_PROC_TYPE_CPU;
}

static inline int _kaapi_get_num_cpus(void)
{
  int cpucount;
  /* compute the number of cpu of the system */
#if defined(__linux__)
  cpucount = sysconf(_SC_NPROCESSORS_CONF);
#elif defined(__APPLE__)
  {
    int mib[2];
    size_t len;
    mib[0] = CTL_HW;
    mib[1] = HW_NCPU;
    len = sizeof(cpucount);
    sysctl(mib, 2, &cpucount, &len, 0, 0);
  }
#elif defined(_WIN32)
  SYSTEM_INFO sys_info;
  GetSystemInfo(&sys_info);
  cpucount = sys_info.dwNumberOfProcessors;
#else
#  warning "Could not compute number of physical cpu of the system. Default value==1"
  cpucount = 1;
#endif
  
  return cpucount;
}

static inline int _kaapi_get_num_gpus(void)
{
  cudaError_t res;
  int devcount;
  
  if ( (res = cudaGetDeviceCount(&devcount)) != cudaSuccess )
  {
    fprintf( stdout, "%s: cudaGetDeviceCount ERROR %d\n",
            __FUNCTION__, res );
    fflush( stdout );
    abort();
  }
  
  return devcount;
}

typedef struct kaapi_calibrate_param_t {
  int src;
  int dest;
} kaapi_calibrate_param_t;

static inline void kaapi_calibrate_host_to_device(const int cpu, const int gpu, size_t size, char* dev, char* const host, cudaStream_t stream)
{
  double t0, t1;
  int i;

  /* latency */  
  t0 = kaapi_get_elapsedus();
  for( i = 0; i < NITER; i++)
  {
    cudaMemcpyAsync(dev, host, sizeof(char), cudaMemcpyHostToDevice, stream);
  }
  cudaStreamSynchronize(stream);
  t1 = kaapi_get_elapsedus();
  kaapi_machine_model[KAAPI_PROC_TYPE_CPU][KAAPI_PROC_TYPE_CUDA].dev[cpu][gpu].latency = (t1-t0)/NITER;

  /* bandwidth */  
  t0 = kaapi_get_elapsedus();
  for( i = 0; i < NITER; i++)
  {
    cudaMemcpyAsync(dev, host, size, cudaMemcpyHostToDevice, stream);
  }
  cudaStreamSynchronize(stream);
  t1 = kaapi_get_elapsedus();
  kaapi_machine_model[KAAPI_PROC_TYPE_CPU][KAAPI_PROC_TYPE_CUDA].dev[cpu][gpu].bandwidth = size/((t1-t0)/NITER);
}

static inline void kaapi_calibrate_device_to_host(const int cpu, const int gpu, size_t size, char* host, char* const dev, cudaStream_t stream)
{
  double t0, t1;
  int i;

  /* latency */
  t0 = kaapi_get_elapsedus();
  for( i = 0; i < NITER; i++)
  {
    cudaMemcpyAsync(host, dev, sizeof(char), cudaMemcpyDeviceToHost, stream);
  }
  cudaStreamSynchronize(stream);
  t1 = kaapi_get_elapsedus();
  kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CPU].dev[gpu][cpu].latency = (t1-t0)/NITER;

  /* bandwidth */
  t0 = kaapi_get_elapsedus();
  for( i = 0; i < NITER; i++)
  {
    cudaMemcpyAsync(host, dev, size, cudaMemcpyDeviceToHost, stream);
  }
  cudaStreamSynchronize(stream);
  t1 = kaapi_get_elapsedus();
  kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CPU].dev[gpu][cpu].bandwidth = size/((t1-t0)/NITER);
}

void* _kaapi_run_calibrate(void* arg)
{
  kaapi_calibrate_param_t* param = (kaapi_calibrate_param_t*)arg;
  const int cpu = param->src;
  const int gpu = param->dest;
  char *dev, *host;
  size_t size;
  
  cudaStream_t stream;
  
  cudaSetDevice(gpu);
  
  cudaFree(0);
  
  cudaStreamCreate(&stream);
  
  size = SIZE;
  
  cudaHostAlloc((void**)&host, size, cudaHostAllocPortable);
  memset(host, 0, size);
  
  cudaMalloc((void**)&dev, size);
  cudaMemset(dev, 0, size);
  
  kaapi_calibrate_host_to_device(cpu, gpu, size, dev, host, stream);
  kaapi_calibrate_device_to_host(cpu, gpu, size, host, dev, stream);
  
  cudaStreamDestroy(stream);
  cudaFree(dev);
  cudaFreeHost(host);
  
  cudaDeviceReset();
  
  return NULL;
}

static inline void kaapi_comm_bandwith_cpu2gpu_run(const int cpu, const int gpu)
{
  pthread_attr_t attr;
  pthread_t tid;
  kaapi_calibrate_param_t* param;
  
  param = (kaapi_calibrate_param_t*)malloc(sizeof(kaapi_calibrate_param_t));
  param->dest = gpu;
  param->src = cpu;
  pthread_attr_init(&attr);
#if !defined(__APPLE__)
  cpu_set_t cpuset;
  CPU_ZERO(&cpuset);
  CPU_SET(cpu, &cpuset);
  pthread_attr_setaffinity_np(&attr, sizeof(cpu_set_t), &cpuset);
  sched_yield();
#endif
  pthread_create(&tid, &attr, &_kaapi_run_calibrate, (void*)param);
  pthread_join(tid, NULL);
  free(param);
}

static inline int kaapi_enable_cuda_peer(int gpu0, int gpu1)
{
  int canAccessPeer;
  
  cudaDeviceCanAccessPeer(&canAccessPeer, gpu0, gpu1);
  if (canAccessPeer) {
    cudaDeviceEnablePeerAccess(gpu1, 0);
    return 1;
  }
  
  return 0;
}

static inline void kaapi_disable_cuda_peer(int gpu0, int gpu1)
{
  cudaDeviceDisablePeerAccess(gpu1);
}

static inline void kaapi_calibrate_device_to_device(size_t size, char* dev0, int gpu0, char* dev1, int gpu1, cudaStream_t stream)
{
  double t0, t1;
  int i;

  /* latency */
  t0 = kaapi_get_elapsedus();
  for( i = 0; i < NITER; i++)
  {
    cudaMemcpyPeerAsync(dev0, gpu0, dev1, gpu1, sizeof(char), stream);
  }
  cudaStreamSynchronize(stream);
  t1 = kaapi_get_elapsedus();
  kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CUDA].dev[gpu1][gpu0].latency = (t1-t0)/NITER;

  /* bandwidth */
  t0 = kaapi_get_elapsedus();
  for( i = 0; i < NITER; i++)
  {
    cudaMemcpyPeerAsync(dev0, gpu0, dev1, gpu1, size, stream);
  }
  cudaStreamSynchronize(stream);
  t1 = kaapi_get_elapsedus();
  kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CUDA].dev[gpu1][gpu0].bandwidth = size/((t1-t0)/NITER);
}

static inline void kaapi_comm_bandwith_gpu2gpu_run(const int gpu0, const int gpu1)
{
  char *dev0, *dev1;
  size_t size = SIZE;
  cudaStream_t stream;

  cudaSetDevice(gpu0);
  if(kaapi_enable_cuda_peer(gpu0, gpu1) == 0)
  {
    cudaDeviceReset();
    return;
  }
  
  /* GPU 1 (src) */
  cudaSetDevice(gpu1);
  cudaFree(0);
  cudaMalloc((void**)&dev1, size);
  cudaMemset(dev1, 1, size);

  /* GPU 0 (dest) */  
  cudaSetDevice(gpu0);
  cudaFree(0);
  cudaMalloc((void**)&dev0, size);
  cudaMemset(dev0, 0, size);
  
  cudaStreamCreate(&stream);
  kaapi_calibrate_device_to_device(size, dev0, gpu0, dev1, gpu1, stream);
  cudaStreamDestroy(stream);
  
  kaapi_disable_cuda_peer(gpu0, gpu1);
  cudaFree(dev0);
  
  /* Free GPU 1 */
  cudaSetDevice(gpu1);
  cudaFree(dev1);
  
  cudaDeviceReset();  
}

static inline void kaapi_perfmodel_comm_bandwidth_calibrate(void)
{
  int cpucount, gpucount;
  int cpu, gpu, gpu1;

  cpucount = _kaapi_get_num_cpus();
  gpucount = _kaapi_get_num_gpus();
  
  for(cpu= 0; cpu < cpucount; cpu++)
  {
    for(gpu= 0; gpu < gpucount; gpu++)
    {
      kaapi_comm_bandwith_cpu2gpu_run(cpu, gpu);
    }
  }
  
  for(gpu= 0; gpu < gpucount; gpu++)
  {
    for(gpu1= 0; gpu1 < gpucount; gpu1++)
    {
      if(gpu != gpu1)
        kaapi_comm_bandwith_gpu2gpu_run(gpu, gpu1);
    }
  }
}

static inline FILE* kaapi_perfmodel_comm_bandwidth_file_open_w(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_COMM_FILE, "w+");
  kaapi_assert(file != NULL);
  return file;
}

static inline void kaapi_perfmodel_comm_bandwidth_file_close(FILE* file)
{
  fclose(file);
}

static inline FILE* kaapi_perfmodel_comm_bandwidth_file_open_r(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_COMM_FILE, "r");
  return file;
}

static inline void kaapi_perfmodel_comm_bandwidth_file_parse(FILE* file)
{
  int ret;
  char src_arch_str[10];
  char dst_arch_str[10];
  int dst, src;
  int dst_type, src_type;
  double latency, bandwidth;
  
  while(1)
  {
    ret = fscanf(file, "%d %s %d %s %le %le\n",
                 &src,
                 src_arch_str,
                 &dst,
                 dst_arch_str,
                 &latency,
                 &bandwidth
                 );
    if((ret == EOF) || (ret != KAAPI_PERFMODEL_MAX_PARAMS))
      break;
    
    src_type = kaapi_get_string_arch(src_arch_str);
    dst_type = kaapi_get_string_arch(dst_arch_str);
    kaapi_assert_debug( (src < KAAPI_MAX_PROCESSOR) && (dst < KAAPI_MAX_PROCESSOR) );
    kaapi_assert_debug( (src_type < KAAPI_PROC_TYPE_MAX) && (dst_type < KAAPI_PROC_TYPE_MAX) );
    kaapi_machine_model[src_type][dst_type].dev[src][dst].latency = latency;
    kaapi_machine_model[src_type][dst_type].dev[src][dst].bandwidth = bandwidth;
  }
}

static inline void kaapi_perfmodel_comm_bandwidth_file_dump(FILE* file)
{
  int src, dst;
  int src_type, dst_type;
  
  for(src_type = 0; src_type < KAAPI_PROC_TYPE_MAX; src_type++) {
    for(dst_type = 0; dst_type < KAAPI_PROC_TYPE_MAX; dst_type++){
      kaapi_pair_entry_t* pair = &kaapi_machine_model[src_type][dst_type];
      
      for(src= 0; src < KAAPI_MAX_PROCESSOR; src++){
        for(dst= 0; dst < KAAPI_MAX_PROCESSOR; dst++){
          if(pair->dev[src][dst].bandwidth == 0.f)
            continue;
          
          // src src_type dest dest_type latency bandwidth
          fprintf(file, "%d %s %d %s %le %le\n",
                  src, kaapi_get_arch_string(src_type),
                  dst, kaapi_get_arch_string(dst_type),
                  pair->dev[src][dst].latency,
                  pair->dev[src][dst].bandwidth
                  );
          fflush(stdout);
        }
      }
    }
  }  
}

static inline void kaapi_perfmodel_comm_bandwidth_print(void)
{
  int src, dst;
  int src_type, dst_type;
  
  for(src_type = 0; src_type < KAAPI_PROC_TYPE_MAX; src_type++) {
    for(dst_type = 0; dst_type < KAAPI_PROC_TYPE_MAX; dst_type++){
      kaapi_pair_entry_t* pair = &kaapi_machine_model[src_type][dst_type];
      
      for(src= 0; src < KAAPI_MAX_PROCESSOR; src++){
        for(dst= 0; dst < KAAPI_MAX_PROCESSOR; dst++){
          if(pair->dev[src][dst].bandwidth == 0.f)
            continue;
          
          // src src_type dest dest_type latency bandwidth
          fprintf(stdout, "%s %d -> %s %d latency=%le bandwidth=%le\n",
                  kaapi_get_arch_string(src_type), src,
                  kaapi_get_arch_string(dst_type), dst,
                  pair->dev[src][dst].latency,
                  pair->dev[src][dst].bandwidth
                  );
          fflush(stdout);
        }
      }
    }
  }
}

void kaapi_perfmodel_comm_bandwidth_init(void)
{
  int src, dst;
  int src_type, dst_type;  
  FILE *file;
  
  for(src_type = 0; src_type < KAAPI_PROC_TYPE_MAX; src_type++) {
    for(dst_type = 0; dst_type < KAAPI_PROC_TYPE_MAX; dst_type++){
      for(src= 0; src < KAAPI_MAX_PROCESSOR; src++){
        for(dst= 0; dst < KAAPI_MAX_PROCESSOR; dst++){
          kaapi_machine_model[src_type][dst_type].dev[src][dst].bandwidth = 0.f;
          kaapi_machine_model[src_type][dst_type].dev[src][dst].latency = 0.f;
        }
      }
    }
  }
  
  file = kaapi_perfmodel_comm_bandwidth_file_open_r();
  if(file != 0)
  {
    kaapi_perfmodel_comm_bandwidth_file_parse(file);
    kaapi_perfmodel_comm_bandwidth_file_close(file);    
  }
  else
  {
    fprintf(stdout, "*** Kaapi: WARNING communication sampling not found for performance model. Calibrating ... ");
    fflush(stdout);
    kaapi_perfmodel_comm_bandwidth_calibrate();
    file = kaapi_perfmodel_comm_bandwidth_file_open_w();
    kaapi_perfmodel_comm_bandwidth_file_dump(file);
    kaapi_perfmodel_comm_bandwidth_file_close(file);
    fprintf(stdout, "done\n");
    fflush(stdout);
  }
}

void kaapi_perfmodel_comm_bandwidth_finalize(void)
{
//  kaapi_perfmodel_comm_bandwidth_print();
}

static inline double kaapi_perfmodel_comm_get_time_device_to_device(kaapi_processor_t* const kdest, kaapi_processor_t* const ksrc, const uint64_t nbytes)
{
  const int src = kaapi_cuda_ctx_get(ksrc)->dev;
  const int dst = kaapi_cuda_ctx_get(kdest)->dev;
  double latency1, bandwidth1;
  double latency2, bandwidth2;
  int host_cpu;
  
  kaapi_assert_debug( (src < KAAPI_MAX_PROCESSOR) && (dst < KAAPI_MAX_PROCESSOR) );
  
  if (kaapi_default_param.cudapeertopeer && (kaapi_cuda_dev_has_peer_access(src, dst)))
  {
    latency1 = kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CUDA].dev[src][dst].latency;
    bandwidth1 = kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CUDA].dev[src][dst].bandwidth;
    return nbytes/(bandwidth1/kaapi_cuda_get_proc_count()) + latency1;
  }
  /* compute time GPU1 -> host -> GPU2 */
  host_cpu = kdest->cpuid;
  latency1 = kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CPU].dev[src][host_cpu].latency;
  bandwidth1 = kaapi_machine_model[KAAPI_PROC_TYPE_CUDA][KAAPI_PROC_TYPE_CPU].dev[src][host_cpu].bandwidth;
  latency2 = kaapi_machine_model[KAAPI_PROC_TYPE_CPU][KAAPI_PROC_TYPE_CUDA].dev[host_cpu][dst].latency;
  bandwidth2 = kaapi_machine_model[KAAPI_PROC_TYPE_CPU][KAAPI_PROC_TYPE_CUDA].dev[host_cpu][dst].bandwidth;
  
  return (nbytes/(bandwidth1/kaapi_cuda_get_proc_count())+ latency1) + (nbytes/(bandwidth2/kaapi_cuda_get_proc_count()) + latency2);
}

static inline int kaapi_perfmodel_comm_bandwidth_get_dev(kaapi_processor_t* const kproc)
{
  if(kaapi_processor_get_type(kproc) == KAAPI_PROC_TYPE_CUDA)
  {
    return kaapi_cuda_ctx_get(kproc)->dev;
  } else {
    return kproc->cpuid;
  }
}

static inline double kaapi_perfmodel_comm_get_time(kaapi_processor_t* const kdest, kaapi_processor_t* const ksrc, const uint64_t nbytes)
{
  const int src_type = kaapi_processor_get_type(ksrc);
  const int dst_type = kaapi_processor_get_type(kdest);
  int src, dst;
  double latency, bandwidth;
  
  kaapi_assert_debug( (src_type < KAAPI_PROC_TYPE_MAX) && (dst_type < KAAPI_PROC_TYPE_MAX) );
  
  if((src_type == KAAPI_PROC_TYPE_CUDA) && (dst_type == KAAPI_PROC_TYPE_CUDA))
    return kaapi_perfmodel_comm_get_time_device_to_device(kdest, ksrc, nbytes);
  
  src = kaapi_perfmodel_comm_bandwidth_get_dev(ksrc);
  dst = kaapi_perfmodel_comm_bandwidth_get_dev(kdest);
  
  kaapi_assert_debug( (src < KAAPI_MAX_PROCESSOR) && (dst < KAAPI_MAX_PROCESSOR) );  
  
  latency = kaapi_machine_model[src_type][dst_type].dev[src][dst].latency;
  bandwidth = kaapi_machine_model[src_type][dst_type].dev[src][dst].bandwidth;

  return nbytes/(bandwidth/kaapi_cuda_get_proc_count()) + latency;
}

double kaapi_perfmodel_comm_bandwidth_predict(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  size_t i;
  void* sp = td->task->sp;
  const size_t count_params = kaapi_format_get_count_params(td->fmt, sp);
  kaapi_address_space_id_t kasid = kaapi_memory_map_kid2asid(kproc->kid);
  uint64_t nbytes = 0;
  double predicted= 0.f, timeus;
  
  for (i = 0; i < count_params; i++) {
    kaapi_access_mode_t m = kaapi_format_get_mode_param(td->fmt, i, sp);
    
    m = KAAPI_ACCESS_GET_MODE(m);
    if (m & KAAPI_ACCESS_MODE_V)
      continue;
    
    kaapi_access_t access = kaapi_format_get_access_param(td->fmt, i, sp);
    kaapi_data_t* const kdata = kaapi_data(kaapi_data_t, &access);
    kaapi_metadata_info_t* const kmdi = kdata->mdi;
    kaapi_assert_debug(kmdi != 0);
    
    if(!kaapi_metadata_info_is_valid(kmdi, kasid))
    {
      nbytes = kaapi_memory_view_size(&kdata->view);
      uint16_t lid = kaapi_metadata_info_first_valid(kmdi);
      kaapi_processor_t* const ksrc =  kaapi_all_kprocessors[kaapi_memory_map_lid2kid(lid)];
      timeus = kaapi_perfmodel_comm_get_time(kproc, ksrc, nbytes);
      if(timeus > 0.f)
      {
        predicted += timeus;
      }
    }
  }
  
  return predicted;
}

#endif /* KAAPI_USE_CUDA */