
#include <stdio.h>
#include <math.h>

#include "kaapi_impl.h"
#include "kaapi_tasklist.h"

#include "kaapi_perfmodel_idgraf.h"

#if defined(KAAPI_USE_CUDA)
#include "machine/cuda/kaapi_cuda_impl.h"
#endif

#include "kaapi_perfmodel_bandwidth.h"
#include "kaapi_perfmodel_history.h"

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

typedef struct kaapi_perfmodel_comm_t {
  void      (*init)       (void);
  void      (*finalize)   (void);
  double    (*predict)    (kaapi_processor_t* const, kaapi_taskdescr_t* const);
  void      (*update)     (const kaapi_address_space_id_t, const kaapi_address_space_id_t, const size_t, const double);
} kaapi_perfmodel_comm_t;

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm;

/**************************************************************************
 null Communication model - always returns zero.
 */
static inline void kaapi_perfmodel_comm_null_init(void)
{
}

static inline void kaapi_perfmodel_comm_null_finalize(void)
{
}

static inline double kaapi_perfmodel_comm_null_predict(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  return 0.f;
}

static inline void kaapi_perfmodel_comm_null_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                      const size_t size, const double time)
{
  
}

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_idgraf = {
  .init                = kaapi_perfmodel_comm_idgraf_init,
  .finalize            = kaapi_perfmodel_comm_idgraf_finalize,
  .predict             = kaapi_perfmodel_comm_idgraf_predict,
  .update              = kaapi_perfmodel_comm_null_update
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_history = {
  .init                = kaapi_perfmodel_comm_history_init,
  .finalize            = kaapi_perfmodel_comm_history_finalize,
  .predict             = kaapi_perfmodel_comm_history_predict,
  .update              = kaapi_perfmodel_comm_history_update_entry
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_bandwidth = {
  .init                = kaapi_perfmodel_comm_bandwidth_init,
  .finalize            = kaapi_perfmodel_comm_bandwidth_finalize,
  .predict             = kaapi_perfmodel_comm_bandwidth_predict,
  .update              = 0
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_null = {
  .init                = kaapi_perfmodel_comm_null_init,
  .finalize            = kaapi_perfmodel_comm_null_finalize,
  .predict             = kaapi_perfmodel_comm_null_predict,
  .update              = kaapi_perfmodel_comm_null_update
};

void kaapi_perfmodel_comm_init(void)
{
  const char* perfmodel = getenv("KAAPI_PERFMODEL_TRANSFER");
  
  kaapi_perfmodel_comm = kaapi_perfmodel_comm_null;
  if( perfmodel != 0 )
  {
    if (strcmp(perfmodel, "idgraf") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_idgraf;
    else if (strcmp(perfmodel, "history") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_history;
    else if (strcmp(perfmodel, "bandwidth") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_bandwidth;
    else if (strcmp(perfmodel, "null") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_null;
    else {
      fprintf(stderr, "*** Kaapi: bad value for 'KAAPI_PERFMODEL_TRANSFER': '%s'\n",
              getenv("KAAPI_PERFMODEL_TRANSFER")
              );
    }
  }
  
  if(kaapi_perfmodel_comm.init != 0)
    kaapi_perfmodel_comm.init();
}

void kaapi_perfmodel_comm_finalize(void)
{
  if(kaapi_perfmodel_comm.finalize != 0)
    kaapi_perfmodel_comm.finalize();
}

double kaapi_perfmodel_comm_predict_transfer(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  if(kaapi_perfmodel_comm.predict != 0)
    return kaapi_perfmodel_comm.predict(kproc, td);
  
  return 0.f;
}

void kaapi_perfmodel_comm_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                        const size_t size, const double time)
{
  if(kaapi_perfmodel_comm.update != 0)
    kaapi_perfmodel_comm.update(src_asid, dst_asid, size, time);
}

