
#include <stdio.h>

#include <float.h>
#include <math.h>

#include "kaapi_impl.h"
#include "perfmodel/kaapi_perfmodel.h"

#define HEFT_MAX_ACTIVATED_TD    256

#define kaapi_max(a,b)  ((a>b)?a:b)

//#define HEFT_VERBOSE

//#define     KAAPI_HEFT_USE_OVERLAP

//#define HEFT_SCHED_PERF

static double perf_all_processors_start[KAAPI_MAX_PROCESSOR]; /* start time */
static double perf_all_processors_end[KAAPI_MAX_PROCESSOR]; /* prediction of end */
static double perf_all_processors_last[KAAPI_MAX_PROCESSOR]; /* tasks on the processor */
static kaapi_lock_t perf_all_processors_lock[KAAPI_MAX_PROCESSOR]; /* lock of value above */

/* executime time factor */
static double kaapi_sched_heft_alpha = 1.0f;

/* communication factor */
static double kaapi_sched_heft_beta = 1.0f;

typedef struct kaapi_sched_heft_task_t {
  kaapi_taskdescr_t*        td;
  double                    speedup;
  double                    tcpu;
  double                    tgpu;
} kaapi_sched_heft_task_t;

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

#if defined(HEFT_SCHED_PERF)
static uint64_t heft_perf_all_processors[KAAPI_MAX_PROCESSOR];

static inline void kaapi_heft_perf_begin(const char* s)
{
  heft_perf_all_processors[kaapi_get_self_kid()] = kaapi_get_elapsedns();
}

static inline void kaapi_heft_perf_end(const char* s, const uint32_t cnt)
{
  const uint64_t tdelta = kaapi_get_elapsedns() - heft_perf_all_processors[kaapi_get_self_kid()];
  fprintf(stdout, "%d:%s: %lu %llu\n", kaapi_get_self_kid(), s, cnt, tdelta);
  fflush(stdout);
}
#endif


static inline void kaapi_heft_init(void)
{
  int i;
  for( i = 0; i < KAAPI_MAX_PROCESSOR; i++)
  {
    perf_all_processors_start[i]= kaapi_get_elapsedus();
    perf_all_processors_end[i] = perf_all_processors_start[i];
    perf_all_processors_last[i] = 0.f;
    kaapi_atomic_initlock(&perf_all_processors_lock[i]);
  }
}

static inline void kaapi_heft_finalize(void)
{
  int i;
  for( i = 0; i < KAAPI_MAX_PROCESSOR; i++)
  {
    perf_all_processors_start[i]= 0.f;
    perf_all_processors_end[i] = 0.f;
    perf_all_processors_last[i] = 0.f;
    kaapi_atomic_destroylock(&perf_all_processors_lock[i]);
  }
}

static inline double kaapi_heft_communication_cost( kaapi_processor_t* const kproc,
                                                            kaapi_taskdescr_t* const td )
{
#if defined(KAAPI_HEFT_USE_OVERLAP)
  const double comm = kaapi_perfmodel_comm_predict_transfer(kproc, td);
  const double overlap = comm - perf_all_processors_last[kproc->kid];
  if(overlap > 0.f)
    return overlap;
  
  return 0.f;
#else
  return kaapi_perfmodel_comm_predict_transfer(kproc, td);
#endif
}

static inline double kaapi_heft_execution_cost(kaapi_processor_t* const kproc,
                                                        kaapi_taskdescr_t* const td )
{
  return kaapi_perfmodel_task_predict(kproc, td);
}

static inline unsigned int kaapi_heft_schedule_rand(kaapi_taskdescr_t* const td)
{
  unsigned int target;
  int arch;
  
  while(1) {
    target = ((unsigned int)rand()) % kaapi_count_kprocessors;
    arch = kaapi_processor_get_type(kaapi_all_kprocessors[target]);    
    if(kaapi_task_has_arch(td->task, arch) && (kaapi_get_task_body_by_arch(td, arch) != 0))
      return target;
  }
}

static inline int kaapi_sched_heft_task_schedule(kaapi_taskdescr_t* const td, int* kid, double* local_perf_task_all_processors,
                                                 double* local_perf_comm_all_processors)
{  
  double perf_predicted_end; /* predicted end of all kprocesssors */
  double perf_end_min;
  double perf_predicted_transfer;
  int i;
  int kid_min;
  int arch;
  
  perf_end_min = DBL_MAX;
  kid_min = kaapi_get_self_kid();
  td->exectime= 0.f;
  
  for(i= 0; i < kaapi_count_kprocessors; i++)
  {
    arch = kaapi_processor_get_type(kaapi_all_kprocessors[i]);
    if(!kaapi_task_has_arch(td->task,arch) || (kaapi_get_task_body_by_arch(td, arch) ==0))
    {
      continue;
    }
    
    local_perf_task_all_processors[i] = kaapi_heft_execution_cost(kaapi_all_kprocessors[i], td);
    local_perf_comm_all_processors[i] = kaapi_heft_communication_cost(kaapi_all_kprocessors[i], td);
    
    if( (fpclassify(local_perf_task_all_processors[i]+local_perf_comm_all_processors[i]) == FP_ZERO) ||
       (isnan(local_perf_task_all_processors[i])) || (isnan(local_perf_comm_all_processors[i])) )
    {
      goto push_random;
    }
    
    perf_predicted_end = kaapi_max(kaapi_get_elapsedus(), perf_all_processors_start[i]) +
      perf_all_processors_last[i] +
      kaapi_sched_heft_alpha * local_perf_task_all_processors[i] +
      kaapi_sched_heft_beta * local_perf_comm_all_processors[i];
    
    /* found best result ? */
    if(perf_predicted_end < perf_end_min)
    {
      kid_min = i;
      perf_end_min = perf_predicted_end;
    }
  }
  
  kaapi_atomic_lock(&perf_all_processors_lock[kid_min]);
  perf_all_processors_start[kid_min] = kaapi_max(kaapi_get_elapsedus(), perf_all_processors_start[kid_min]);
  perf_all_processors_end[kid_min] = perf_all_processors_start[kid_min] + perf_all_processors_last[kid_min];
  
  perf_all_processors_end[kid_min] += local_perf_task_all_processors[kid_min];
  perf_all_processors_last[kid_min] += local_perf_task_all_processors[kid_min];
  
  if((kaapi_get_elapsedus() + local_perf_comm_all_processors[kid_min]) < perf_all_processors_end[kid_min])
  {
    perf_predicted_transfer = 0.f;
  }
  else
  {
    perf_predicted_transfer = (kaapi_get_elapsedus() + local_perf_comm_all_processors[kid_min]) - perf_all_processors_end[kid_min];
  }
  perf_all_processors_end[kid_min] += perf_predicted_transfer;
  perf_all_processors_last[kid_min] += perf_predicted_transfer;  
  kaapi_atomic_unlock(&perf_all_processors_lock[kid_min]);
  
  td->exectime = local_perf_task_all_processors[kid_min];
  td->predicted_transfer = perf_predicted_transfer;
  
#if defined(KAAPI_DEBUG)
  td->endtime = perf_end_min;
  td->kid = kid_min;
#endif
  *kid = kid_min;
  
#if defined(HEFT_VERBOSE)
  fprintf(stdout,"[%s] td=%p(task=%p) best %d -> %d (%s) predicted=%.10f start=%.10f end=%.10f last=%.10f\n", __FUNCTION__,
          (void*)td,
          (void*)td->task,
          kaapi_get_self_kid(),
          kid_min,
          td->fmt->name,
          perf_end_min,
          perf_all_processors_start[kid_min],
          perf_all_processors_end[kid_min],
          perf_all_processors_last[kid_min]
          );
  fflush(stdout);
#endif
  return 0;
  
  /* in case we can not predict execution for all processors,
     pick a random processor able to execute it */
push_random:
  kid_min = kaapi_heft_schedule_rand(td);
  *kid = kid_min;
  td->exectime = 0.f;
  td->predicted_transfer = 0.f;
#if defined(KAAPI_DEBUG)
  td->endtime = 0.f;
  td->kid = kid_min;
#endif
  kaapi_atomic_lock(&perf_all_processors_lock[kid_min]);
  perf_all_processors_start[kid_min] = kaapi_max(kaapi_get_elapsedus(), perf_all_processors_start[kid_min]);
  perf_all_processors_end[kid_min] = perf_all_processors_start[kid_min]+perf_all_processors_last[kid_min];
  kaapi_atomic_unlock(&perf_all_processors_lock[kid_min]);
  /* We can not predict, so null */
#if defined(HEFT_VERBOSE)
  fprintf(stdout,"[%s] td=%p(task=%p) random -> %d  predicted=%.10f (%s)\n", __FUNCTION__,
          (void*)td,
          (void*)td->task,
          kid_min,
          local_perf_task_all_processors[kid_min],
          td->fmt->name
          );
  fflush(stdout);
#endif
  
  return 0;
}

static inline void kaapi_heft_prologue(kaapi_processor_t* const kproc,
                                  kaapi_taskdescr_t* const td)
{
#if defined(KAAPI_DEBUG)
  if(td->kid >= 0)
  {
    kaapi_assert_debug(kproc->kid == td->kid);
  }
#endif
  kaapi_atomic_lock(&perf_all_processors_lock[kproc->kid]);
  perf_all_processors_last[kproc->kid] -= (td->exectime + td->predicted_transfer);
  perf_all_processors_start[kproc->kid] = kaapi_get_elapsedus() + td->exectime; /* only considers kernel execution */
  perf_all_processors_end[kproc->kid] = perf_all_processors_start[kproc->kid]+perf_all_processors_last[kproc->kid];
  kaapi_atomic_unlock(&perf_all_processors_lock[kproc->kid]);
}

static inline void kaapi_heft_epilogue(kaapi_processor_t* const kproc,
                                  kaapi_taskdescr_t* const td)
{
#if defined(HEFT_VERBOSE)
  kaapi_atomic_lock(&perf_all_processors_lock[kproc->kid]);
  fprintf(stdout,"[%s] td=%p (%s) kid=%d target=%d start=%.2f end=%.2f last=%.2f predicted=%.2f end=%.2f (err=%.2f)\n",
          __FUNCTION__,
          (void*)td,
          td->fmt->name,
          kproc->kid,
          td->kid,
          perf_all_processors_start[kproc->kid],
          perf_all_processors_end[kproc->kid],
          perf_all_processors_last[kproc->kid],
          td->exectime,
          td->endtime,
          (kaapi_get_elapsedus()-td->endtime)
          );
  fflush(stdout);
  kaapi_atomic_unlock(&perf_all_processors_lock[kproc->kid]);
#endif
  /* */
}

/** Push one ready task to the correct queue.
 * 0) it tests the site is set
 * 1) it tests the task priority and
 * 2) affinity (if enabled)
 * 3) them pushes in the readylist rtl
 */
static inline int kaapi_sched_heft_pushready_td(kaapi_readytasklist_t*  rtl, kaapi_taskdescr_t*      td)
{
  kaapi_processor_t* curr_kproc = kaapi_get_current_processor();
  
  if(td->fmt == 0)
  {
    return kaapi_readylist_push( rtl, td );
  }
  
  if (kaapi_taskdescr_get_site(td) !=0)
  {
    return kaapi_sched_push_by_site(td);
  }
  
  if ( (kaapi_get_task_bodywh_by_arch(td, kaapi_processor_get_type(curr_kproc)) == 0) ||
      (kaapi_task_has_arch(td->task, kaapi_processor_get_type(curr_kproc)) == 0) )
  {
    kaapi_assert_debug( td->tasklist != 0 );
#if defined(KAAPI_DEBUG)
    td->kid = -1;
#endif
    return kaapi_readylist_push( &td->tasklist->rtl, td );
  }
  
  return kaapi_readylist_push( rtl, td );
}

static inline int kaapi_sched_heft_push_from_activationlist(kaapi_readytasklist_t* const rtl, kaapi_activationlink_t*	head)
{
  kaapi_taskdescr_t* td;
  int retval =0;
  
  while (head !=0) {
    td = head->td;
    if (kaapi_taskdescr_activated(td)) {
      ++retval;
      kaapi_sched_heft_pushready_td(rtl, td);
    }
    head = head->next;
  }
  return retval;
}

static int kaapi_sched_task_cmp(const void *a, const void *b)
{
  return (((kaapi_sched_heft_task_t*)b)->speedup - ((kaapi_sched_heft_task_t*)a)->speedup);
}

/**
 Here it activate our depedencies and calculates the speedup (in respect with the GPU).
 Then, it sorts the array of ready tasks by decreasing speedup. This will be our scheduling order later.
 */
static inline int kaapi_sched_heft_activate_list(kaapi_sched_heft_task_t* const td_activated, kaapi_activationlist_t* const list)
{
  kaapi_taskdescr_t* td;
  kaapi_activationlink_t* head = list->front;
  int i =0;
  
  while (head !=0) {
    td = head->td;
    if (kaapi_taskdescr_activated(td)) {
      td_activated[i].td = td;
      td_activated[i].tcpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_HOST, head->td);
      td_activated[i].tgpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_CUDA, head->td);
      td_activated[i].speedup = DBL_MAX;
      if((fpclassify(td_activated[i].tcpu) != FP_ZERO) && (fpclassify(td_activated[i].tgpu) != FP_ZERO))
      {
        td_activated[i].speedup = td_activated[i].tcpu / td_activated[i].tgpu;
      }
      i++;
    }
    head = head->next;
  }
  
  if(i > 1)
  {
    /* quicksort tasks by speedup */
    qsort(td_activated, i, sizeof(kaapi_sched_heft_task_t), kaapi_sched_task_cmp);
    
#if defined(GDUAL_VERBOSE)
    fprintf(stdout,"%d:%s: first=%.4f last=%.4f n=%d\n", kaapi_get_self_kid(), __FUNCTION__,
            td_activated[0].speedup, td_activated[i-1].speedup, i);
    fflush(stdout);
#endif
  }
  td_activated[i].td = 0;
  
  return i;
}

static inline int kaapi_sched_heft_remotepush_td(kaapi_readytasklist_t*  rtl, const int kid, kaapi_taskdescr_t*  td)
{
  if(td->fmt == 0)
  {
		/* move/alloc task */
    return kaapi_readylist_push( rtl, td );
  }
  
  if (kaapi_taskdescr_get_site(td) !=0)
  {
		/* this task already have a destination */
#if defined(HEFT_VERBOSE)
    fprintf(stdout, "%d:%s: td=%p SITE site=%d\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, kaapi_taskdescr_get_site(td));
    fflush(stdout);
#endif
    return kaapi_sched_push_by_site(td);
  }
  
#if 0
  if(kaapi_default_param.noremotesteal)
  {
		/* disable the steal of this task */
    kaapi_task_set_unstealable(td->task);
  }
#endif
  
  if(kid == kaapi_get_self_kid())
  {
		/* local push */
#if defined(HEFT_VERBOSE)
    fprintf(stdout, "%d:%s: td=%p LOCAL\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td);
    fflush(stdout);
#endif
    return kaapi_readylist_push( kaapi_get_current_processor()->rtl, td );
		/* TODO */
#if 0
    return kaapi_readylist_push( rtl, td );
#endif
  }
  
#if defined(HEFT_VERBOSE)
  fprintf(stdout, "%d:%s: td=%p REMOTE kid=%d\n", kaapi_get_self_kid(), __FUNCTION__,
          (void*)td, kid);
  fflush(stdout);
#endif
  
	/* HERE we push a task to the remote resource */
  return kaapi_readylist_remote_push( kaapi_all_kprocessors[kid]->rtl, td );
}

static inline int kaapi_sched_heft_schedule_update_and_push(
                                                             kaapi_sched_heft_task_t* const td_activated,
                                                             kaapi_readytasklist_t* const rtl,
                                                             int* scheduling)
{
  int i;
  
  i = 0;
  while (td_activated[i].td != 0) {
		/* push a tasks to the remote processor (if needed) */
    kaapi_sched_heft_remotepush_td(rtl, scheduling[i], td_activated[i].td);
    i++;
  }
  
  return 0;
}

static inline int kaapi_sched_heft_schedule_all_tasks(kaapi_sched_heft_task_t* const td_activated,
                                                       double* local_perf_task_all_processors,
                                                       double* local_perf_comm_all_processors,
                                                       int* scheduling)
{
  int i;
  i = 0;
  
  while(td_activated[i].td != 0)
  {
    kaapi_sched_heft_task_schedule(td_activated[i].td, &scheduling[i], local_perf_task_all_processors, local_perf_comm_all_processors);
    i++;
  }
  
#if defined(HEFT_VERBOSE)
  fprintf(stdout, "%d:%s: count=%d\n", kaapi_get_self_kid(), __FUNCTION__, i);
  fflush(stdout);
#endif
  return 0;
}

/**
 It schedules tasks over GPUs and then over CPUs using a lambda until the error (GDUAL_ERROR).
 */
static inline int kaapi_sched_heft_schedule(kaapi_readytasklist_t* const rtl, kaapi_sched_heft_task_t* const td_activated, uint32_t count)
{
  double local_perf_task_all_processors[KAAPI_MAX_PROCESSOR]; /* used to calculate execution time of tasks */
  double local_perf_comm_all_processors[KAAPI_MAX_PROCESSOR]; /* used to calculate transfer time of tasks */
  int sched_kproc[HEFT_MAX_ACTIVATED_TD]; /* destination (kid) of each task */
  
  memset(sched_kproc, 0, HEFT_MAX_ACTIVATED_TD*sizeof(int));
#if defined(HEFT_VERBOSE)
  fprintf(stdout, "%d:%s: BEGIN count=%d lower=%.4f upper=%.4f\n", kaapi_get_self_kid(), __FUNCTION__, count, lower_bound, upper_bound);
  fflush(stdout);
#endif
  kaapi_sched_heft_schedule_all_tasks(td_activated, local_perf_task_all_processors, local_perf_comm_all_processors, sched_kproc);
  
  /* update execution times for each processor (global operation) and push tasks */
  kaapi_sched_heft_schedule_update_and_push(td_activated, rtl, sched_kproc);
  
  return 0;
}

static inline uint32_t kaapi_sched_heft_pushactivated(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  uint32_t cnt_pushed= 0;
  kaapi_sched_heft_task_t td_activated[HEFT_MAX_ACTIVATED_TD];
  
  /* push in the front the activated tasks */
  if (!kaapi_activationlist_isempty(&td->u.acl.list))
  {
#if defined(HEFT_SCHED_PERF)
    kaapi_heft_perf_begin(__FUNCTION__);
#endif
//    cnt_pushed = kaapi_sched_heft_push_from_activationlist( rtl, td->u.acl.list.front );
		/* activates tasks and generates our array of ready tasks (td_activated) */
    cnt_pushed = kaapi_sched_heft_activate_list(td_activated, &td->u.acl.list);
		/* schedules tasks by remote push */
    kaapi_sched_heft_schedule( rtl, td_activated, cnt_pushed );
#if defined(HEFT_SCHED_PERF)
    kaapi_heft_perf_end(__FUNCTION__, cnt_pushed);
#endif
  }
  else
    cnt_pushed = 0;
  
  /* do bcast after child execution (they can produce output data) */
  if (td->u.acl.bcast !=0)
    cnt_pushed += kaapi_sched_heft_push_from_activationlist( rtl, td->u.acl.bcast->front );
  
  return cnt_pushed;
}

/* update counters of thief and victim */
static inline void kaapi_sched_heft_steal_prologue(kaapi_processor_t* const victim, kaapi_taskdescr_t* td)
{
  kaapi_processor_t* const kproc = kaapi_get_current_processor();
  /* re-calculate execution costs for the thief */
  double local_predicted_exectime = kaapi_heft_execution_cost(kproc, td);
  double local_predicted_transfer = kaapi_heft_communication_cost(kproc, td);
  
#if 0
  fprintf(stdout, "%d:%s: STEAL victim=%d td=%p (%s) oldexectime=%.4f newexectime=%.4f\n",
          kaapi_get_self_kid(), __FUNCTION__, victim->kid, (void*)td, td->fmt->name,
          td->exectime, local_perf_task);
  fflush(stdout);
#endif  
  
  /* first update counters on the victim */  
  kaapi_atomic_lock(&perf_all_processors_lock[victim->kid]);
  perf_all_processors_start[victim->kid] = kaapi_max(kaapi_get_elapsedus(), perf_all_processors_start[victim->kid]);
  perf_all_processors_last[victim->kid] -= (td->exectime + td->predicted_transfer);
  perf_all_processors_end[victim->kid] = perf_all_processors_start[victim->kid] + perf_all_processors_last[victim->kid];
  kaapi_atomic_unlock(&perf_all_processors_lock[victim->kid]);
  
  /* then update counters on the thief with its execution cost */  
  kaapi_atomic_lock(&perf_all_processors_lock[kproc->kid]);
  perf_all_processors_start[kproc->kid] = kaapi_max(kaapi_get_elapsedus(), perf_all_processors_start[kproc->kid]);
  perf_all_processors_end[kproc->kid] = perf_all_processors_start[kproc->kid] + perf_all_processors_last[kproc->kid];
  
  perf_all_processors_end[kproc->kid] += local_predicted_exectime;
  perf_all_processors_last[kproc->kid] += local_predicted_exectime;
  
  if((kaapi_get_elapsedus() + local_predicted_transfer) < perf_all_processors_end[kproc->kid])
  {
    local_predicted_transfer = 0.f;
  }
  else
  {
    local_predicted_transfer = (kaapi_get_elapsedus() + local_predicted_transfer) - perf_all_processors_end[kproc->kid];
  }
  perf_all_processors_end[kproc->kid] += local_predicted_transfer;
  perf_all_processors_last[kproc->kid] += local_predicted_transfer;
  kaapi_atomic_unlock(&perf_all_processors_lock[kproc->kid]);
  
  td->exectime = local_predicted_exectime;
  td->predicted_transfer = local_predicted_transfer;
}

/* return the first td that can be executed on arch when steal is enabled for HEFT
 */
static inline kaapi_taskdescr_t* kaapi_sched_heft_steal_first(kaapi_processor_t* const thief, kaapi_taskdescr_t* const first)
{
  int arch = kaapi_processor_get_type(kaapi_get_current_processor());
  kaapi_taskdescr_t* td = first;
  //  int arch = kaapi_processor_get_type(thief);
  
  /* HEFT only steals from the same architecture */
  if(arch != kaapi_processor_get_type(thief))
  {
    return 0;
  }
  
  /* only steal for the righ processor arch or if fmt ==0 (means internal task) */
  /* TODO
   (kaapi_task_getbody(td->task) != (kaapi_task_body_t)kaapi_staticschedtask_body)
   */
  while ((td != 0) &&
         (kaapi_task_is_unstealable(td->task) || !kaapi_task_has_arch(td->task, arch) || ((td->fmt != 0) && (kaapi_get_task_body_by_arch(td, arch) == 0)))
         )
  {
    td = td->prev;
  }
#if 0
  if(td != 0 && td->fmt != 0)
  {
    fprintf(stdout, "%d:%s: td=%p (%s) STEAL (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, td->fmt->name, thief->kid);
    fflush(stdout);
  }
#endif
  
  if((td != 0) && (td->fmt != 0))
  {
    /* we need to update counters from victim and thief */
    kaapi_sched_heft_steal_prologue(thief, td);
  }
  
  return td;
}


void kaapi_sched_heft_init(void)
{
  const char* alpha = getenv("KAAPI_HEFT_ALPHA");
  const char* beta = getenv("KAAPI_HEFT_BETA");
  
  if(alpha != 0)
  {
    kaapi_sched_heft_alpha = atof(alpha);
  }
  if(beta != 0)
  {
    kaapi_sched_heft_beta = atof(beta);
  }
  
  /* enable steal of pushed tasks */
  if (getenv("KAAPI_HEFT_STEAL") !=0)
  {
    kaapi_default_param.noremotesteal = 0;
  }
  else
  {
    kaapi_default_param.noremotesteal = 1;
  }
  kaapi_default_param.perfmodel = 1;
  kaapi_perfmodel_init(); 
  kaapi_heft_init();
}

void kaapi_sched_heft_finalize(void)
{
  if(kaapi_default_param.perfmodel)
  {
    kaapi_perfmodel_finalize();
  }
  kaapi_default_param.perfmodel = 0;
  kaapi_default_param.noremotesteal = 0;
  kaapi_heft_finalize();
}

kaapi_taskdescr_t* kaapi_sched_heft_steal(kaapi_processor_t* const thief, kaapi_taskdescr_t* const td)
{
  if(kaapi_default_param.noremotesteal == 0)
  {
    return kaapi_sched_heft_steal_first(thief, td);
  }
  
  return 0;
}

uint32_t kaapi_sched_heft_push_activated( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  return kaapi_sched_heft_pushactivated( &tasklist->rtl, td );
}

void kaapi_sched_heft_push( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  kaapi_sched_heft_pushready_td(&tasklist->rtl, td);
}

void kaapi_sched_heft_task_epilogue(kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_epilogue(kaapi_get_current_processor(), td);  
  kaapi_heft_epilogue(kaapi_get_current_processor(), td);
}

void kaapi_sched_heft_task_prologue(kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_prologue(kaapi_get_current_processor(), td);  
  kaapi_heft_prologue(kaapi_get_current_processor(), td);
}
