
#include <stdio.h>
#include <float.h>
#include <math.h>

#include "kaapi_impl.h"
#include "perfmodel/kaapi_perfmodel.h"

#define SPEEDUP_MAX_ACTIVATED_TD    256

static double kaapi_sched_speedup_threshold_cpu = 1.0f;
static double kaapi_sched_speedup_threshold_cuda = 5.0f;

typedef struct kaapi_sched_speedup_task_t {
  kaapi_taskdescr_t*        td;
  double                    speedup;
  double                    tcpu;
  double                    tgpu;
} kaapi_sched_speedup_task_t;

static inline int kaapi_sched_speedup_pushready_td(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  kaapi_processor_t* curr_kproc = kaapi_get_current_processor();
  
  if(td->fmt == 0)
  {
    return kaapi_readylist_push( rtl, td );
  }
  
  if (kaapi_taskdescr_get_site(td) !=0)
  {
    return kaapi_sched_push_by_site(td);
  }
  
#if defined(KAAPI_USE_CUDA)
  /* optimize CUDA tasks if we are in a GPU thread (KAAPI_PROC_TYPE_CUDA) and
   it does not have a CPU body (KAAPI_PROC_TYPE_CPU). This line optmize DGEMM almost 2x.
   */
  if (kaapi_processor_get_type(curr_kproc) == KAAPI_PROC_TYPE_CUDA)
  {
    if ( (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CPU) == 0) ||
        (kaapi_task_has_arch(td->task, KAAPI_PROC_TYPE_CPU) == 0) )
    {
      return kaapi_readylist_push( curr_kproc->rtl, td );
    }
  }
#endif
  
  /* for test only: push on master list task with highest priority */
  if ((td->priority == KAAPI_TASKLIST_MAX_PRIORITY) && (curr_kproc->kid != 0))
    return kaapi_readylist_remote_push( rtl, td );
  
  return kaapi_readylist_push( rtl, td );
}

static int kaapi_sched_task_cmp(const void *a, const void *b)
{
  return (((kaapi_sched_speedup_task_t*)b)->speedup - ((kaapi_sched_speedup_task_t*)a)->speedup);
}

/**
 Here it activate our depedencies and calculates the speedup (in respect with the GPU).
 Then, it sorts the array of ready tasks by decreasing speedup. This will be our scheduling order later.
 */
static inline int kaapi_sched_speedup_activate_list(kaapi_sched_speedup_task_t* const td_activated, kaapi_activationlist_t* const list)
{
  kaapi_taskdescr_t* td;
  kaapi_activationlink_t* head = list->front;
  int i =0;
  
  while (head !=0) {
    td = head->td;
    if (kaapi_taskdescr_activated(td)) {
      td_activated[i].td = td;
      td_activated[i].tcpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_HOST, head->td);
      td_activated[i].tgpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_CUDA, head->td);
      td_activated[i].speedup = 0.f;
      if((fpclassify(td_activated[i].tcpu) != FP_ZERO) && (fpclassify(td_activated[i].tgpu) != FP_ZERO))
      {
        td_activated[i].speedup = td_activated[i].tcpu / td_activated[i].tgpu;
#if defined(SPEEDUP_VERBOSE)
        fprintf(stdout,"%d:%s: td=%p(%s) speedup=%.4f n=%d\n", kaapi_get_self_kid(), __FUNCTION__,
                (void*)td_activated[i].td, td_activated[i].td->fmt->name,
                td_activated[i].speedup, i);
        fflush(stdout);
#endif        
      }
      i++;
    }
    head = head->next;
  }
  
  if(i > 1)
  {
    /* quicksort tasks by speedup */
    qsort(td_activated, i, sizeof(kaapi_sched_speedup_task_t), kaapi_sched_task_cmp);
  }
  td_activated[i].td = 0;
  
  return i;
}

static inline int kaapi_sched_speedup_remotepush_td(kaapi_readytasklist_t*  rtl, kaapi_sched_speedup_task_t* const td_activated)
{
  kaapi_processor_t* curr_kproc = kaapi_get_current_processor();
  kaapi_taskdescr_t* td = td_activated->td;
  
  if(td->fmt == 0)
  {
    return kaapi_readylist_push( rtl, td );
  }
  
  if (kaapi_taskdescr_get_site(td) !=0)
  {
    return kaapi_sched_push_by_site(td);
  }
  
#if defined(KAAPI_USE_CUDA)
  if(fpclassify(td_activated->speedup) != FP_ZERO)
  {
    if((td_activated->speedup < kaapi_sched_speedup_threshold_cpu) &&
       (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CPU) != 0))
    {
      /* force to be CPU task if speedup is insignificant */
      kaapi_task_set_arch(td->task, KAAPI_PROC_TYPE_CPU);
    }
    if((td_activated->speedup > kaapi_sched_speedup_threshold_cuda) &&
       (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CUDA) != 0))
    {
      /* force to be CUDA task if speedup is significant */
      kaapi_task_set_arch(td->task, KAAPI_PROC_TYPE_CUDA);
    }
  }
  
  if ( (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CPU) == 0) ||
      (kaapi_task_has_arch(td->task, KAAPI_PROC_TYPE_CPU) == 0) )
  {
    return kaapi_readylist_push( curr_kproc->rtl, td );
  }
#endif
  
  /* for test only: push on master list task with highest priority */
  if ((td->priority == KAAPI_TASKLIST_MAX_PRIORITY) && (curr_kproc->kid != 0))
    return kaapi_readylist_remote_push( rtl, td );
  
  return kaapi_readylist_push( rtl, td );
}

static inline int kaapi_sched_speedup_schedule(kaapi_readytasklist_t* const rtl, kaapi_sched_speedup_task_t* const td_activated, uint32_t count)
{
  int i;
  
  i = 0;
  while (td_activated[i].td != 0) {
		/* push a tasks to the remote processor (if needed) */
    kaapi_sched_speedup_remotepush_td(rtl, &td_activated[i]);
    i++;
  }
  
  return 0;
}

static inline uint32_t kaapi_sched_speedup_pushactivated(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  uint32_t cnt_pushed= 0;
  kaapi_sched_speedup_task_t td_activated[SPEEDUP_MAX_ACTIVATED_TD];
  
  /* push in the front the activated tasks */
  if (!kaapi_activationlist_isempty(&td->u.acl.list))
  {
    cnt_pushed = kaapi_sched_speedup_activate_list(td_activated, &td->u.acl.list);
    kaapi_sched_speedup_schedule( rtl, td_activated, cnt_pushed );
  }
  else
    cnt_pushed = 0;
  
  /* do bcast after child execution (they can produce output data) */
  if (td->u.acl.bcast !=0)
  {
    cnt_pushed += kaapi_sched_speedup_activate_list(td_activated, td->u.acl.bcast);
    kaapi_sched_speedup_schedule( rtl, td_activated, cnt_pushed );
  }
  
  return cnt_pushed;
}

/* return the first td that can be executed on arch
 */
static inline kaapi_taskdescr_t* kaapi_sched_speedup_steal_first(kaapi_processor_t* const thief, kaapi_taskdescr_t* const first)
{
  int arch = kaapi_processor_get_type(kaapi_get_current_processor());
  kaapi_taskdescr_t* td = first;
  //  int arch = kaapi_processor_get_type(thief);
  
  /* only steal for the righ processor arch or if fmt ==0 (means internal task) */
  /* TODO
   (kaapi_task_getbody(td->task) != (kaapi_task_body_t)kaapi_staticschedtask_body)
   */
  while ((td != 0) &&
         (kaapi_task_is_unstealable(td->task) || !kaapi_task_has_arch(td->task, arch) || ((td->fmt != 0) && (kaapi_get_task_body_by_arch(td, arch) == 0)))
         )
  {
    td = td->prev;
  }
#if 0
  if(td != 0 && td->fmt != 0)
  {
    fprintf(stdout, "%d:%s: td=%p (%s) STEAL (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, td->fmt->name, thief->kid);
    fflush(stdout);
  }
#endif
  return td;
}

static inline void kaapi_sched_speedup_compute_speedup(kaapi_sched_speedup_task_t* const td_ready, kaapi_taskdescr_t* const td)
{
  td_ready->td = td;
  td_ready->tcpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_HOST, td);
  td_ready->tgpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_CUDA, td);
  td_ready->speedup = 0.f;
  if((fpclassify(td_ready->tcpu) != FP_ZERO) && (fpclassify(td_ready->tgpu) != FP_ZERO))
  {
    td_ready->speedup = td_ready->tcpu / td_ready->tgpu;
#if defined(SPEEDUP_VERBOSE)
    fprintf(stdout,"%d:%s: td=%p(%s) speedup=%.4f\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td_ready->td, td_ready->td->fmt->name,
            td_ready->speedup);
    fflush(stdout);
#endif
  }
}

static int kaapi_sched_task_cmp_cpu(const void *a, const void *b)
{
  return (((kaapi_sched_speedup_task_t*)a)->speedup - ((kaapi_sched_speedup_task_t*)b)->speedup);
}

/* return the first td that can be executed on arch
 */
static inline kaapi_taskdescr_t* kaapi_sched_speedup_steal_by_cost(kaapi_processor_t* const thief, kaapi_taskdescr_t* const first)
{
  int arch = kaapi_processor_get_type(kaapi_get_current_processor());
  kaapi_taskdescr_t* td = first;
  kaapi_sched_speedup_task_t td_ready[SPEEDUP_MAX_ACTIVATED_TD];
  int i= 0;
  
  while(td != 0)
  {
//    if(kaapi_task_is_unstealable(td->task) || !kaapi_task_has_arch(td->task, arch) || ((td->fmt != 0) && (kaapi_get_task_body_by_arch(td, arch) == 0)))
    if( (td->fmt != 0) && (!kaapi_task_is_unstealable(td->task)) && kaapi_task_has_arch(td->task, arch) && kaapi_get_task_body_by_arch(td, arch))    
    {
      kaapi_sched_speedup_compute_speedup(&td_ready[i], td);
      i++;
    }
    td = td->prev;
  }

  if(i == 0)
  {
    return first;
  }

  td_ready[i].td = 0;  
  if(i > 1)
  {
    if(arch == KAAPI_PROC_TYPE_CPU)
      qsort(td_ready, i, sizeof(kaapi_sched_speedup_task_t), kaapi_sched_task_cmp_cpu);
    else
      qsort(td_ready, i, sizeof(kaapi_sched_speedup_task_t), kaapi_sched_task_cmp);
#if 0
    td = td_ready[0].td;
    if(td != 0 && td->fmt != 0)
    {
      fprintf(stdout, "%d:%s: td=%p (%s) speedup=%.4f STEAL (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
              (void*)td, td->fmt->name, td_ready[0].speedup, thief->kid);
      fprintf(stdout, "%d:%s: td=%p (%s) speedup=%.4f NOT (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
              (void*)td_ready[1].td, td_ready[1].td->fmt->name, td_ready[1].speedup, thief->kid);
      
      fflush(stdout);
    }
#endif
    return td_ready[0].td;
  }

  /* else return the first task */
  return first;
}

void kaapi_sched_speedup_init(void)
{
  const char* speedup_cpu = getenv("KAAPI_SPEEDUP_THRESHOLD_CPU");
  const char* speedup_cuda = getenv("KAAPI_SPEEDUP_THRESHOLD_CUDA");

  if(speedup_cpu != 0)
  {
    kaapi_sched_speedup_threshold_cpu = atof(speedup_cpu);
  }
  if(speedup_cuda != 0)
  {
    kaapi_sched_speedup_threshold_cuda = atof(speedup_cuda);
  }
  
  kaapi_default_param.perfmodel = 1;
  kaapi_perfmodel_init();  
}

void kaapi_sched_speedup_finalize(void)
{
  if(kaapi_default_param.perfmodel)
  {
    kaapi_perfmodel_finalize();
  }
  kaapi_default_param.perfmodel = 0;
}

uint32_t kaapi_sched_speedup_push_activated( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  return 0;
//  return kaapi_sched_speedup_pushactivated( &tasklist->rtl, td );
}

void kaapi_sched_speedup_push( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  kaapi_sched_speedup_pushready_td(&tasklist->rtl, td);
}

kaapi_taskdescr_t* kaapi_sched_speedup_steal(kaapi_processor_t* const thief, kaapi_taskdescr_t* const td)
{
  return kaapi_sched_speedup_steal_by_cost(thief, td);
}

void kaapi_sched_speedup_task_epilogue(kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_epilogue(kaapi_get_current_processor(), td);    
}

void kaapi_sched_speedup_task_prologue(kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_prologue(kaapi_get_current_processor(), td);
}
