/*
** kaapi_hws_pushtask.c
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com / fabien.lementec@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#include "kaapi_impl.h"


/* Push the task thread->pc on a given processor
   Here the method is to use a specific mailbox for kprocessor that 
   contains only task object simple linked in a LIFO order.
   A FIFO order would better allows concurrency between thief and victim.
   The lock on the kproc of the remote mail box is taken in this function. 
   It is also locked on steal operation, see stealreadytasklist.
 */
int kaapi_thread_remotepush_task (
  kaapi_thread_t*      thread,
  kaapi_hws_levelid_t  levelid,
  kaapi_ldid_t         id,
  int                  flag_strict
)
{
  kaapi_assert_debug((flag_strict == KAAPI_TASK_S_DEFAULT) || (flag_strict == KAAPI_TASK_S_STRICT));

  /* the task is in the own queue of the thread but not yet visible (pushed) to other threads */
  kaapi_task_t* local_task = kaapi_thread_toptask(thread);
  
  kaapi_queue_t* queue = kaapi_queue_getperlevel( levelid, id );
  if (queue ==0) 
    return EINVAL;
  
  //if (1) // BUG HERE: non strict is incorrect (flag_strict == KAAPI_TASK_S_STRICT)
  {
    /* push into the remote queue as well as in the local queue:
       - create a new steal_body task as if the local task has been theft,
       so that the owner thread will block until terminaison
       - push the steal_body task into the remote queue
    */
    kaapi_task_t* remote_task    =
        (kaapi_task_t*)kaapi_thread_pushdata_align(
                            thread, 
                            sizeof(kaapi_task_t), 
                            sizeof(uintptr_t)
        );
    
    kaapi_tasksteal_arg_t* argsteal = 
        (kaapi_tasksteal_arg_t*)kaapi_thread_pushdata_align(
                            thread, 
                            sizeof(kaapi_tasksteal_arg_t), 
                            sizeof(void*)
        );
    argsteal->origin_task           = local_task;
    argsteal->origin_body           = local_task->body;
    argsteal->origin_fmt            = 0;    /* should be computed by a thief */
    argsteal->war_param             = 0;    /* assume no war, else to do */
    argsteal->cw_param              = 0;    /* assume no cw mode, else to do */
    kaapi_task_init(
        remote_task, 
        kaapi_tasksteal_body, 
        argsteal
    );

    /* recopy the flag of the local task, initialise the preemption field */
    remote_task->u.s.flag      = local_task->u.s.flag;
    kaapi_task_unset_strictaffinity( remote_task );
    if (flag_strict == KAAPI_TASK_S_STRICT)
      kaapi_task_set_strictaffinity( remote_task );
    local_task->reserved       = remote_task;
    
    /* mark steal should always succeed */
    argsteal->origin_body = kaapi_task_marksteal( local_task ) ? local_task->body : 0;
    kaapi_assert_debug( argsteal->origin_body != 0);

    /* push the remote task on remote queue:
       TODO: select which remote queue (local / kproc or numa)
    */
    kaapi_sched_lock( &queue->lock );
//printf("Remote push: task: %p on queue (id:%i, lid:%i)\n", (void*)remote_task, queue->id, queue->lid );
    kaapi_queue_push( queue, remote_task);
    kaapi_sched_unlock( &queue->lock );

    /* ok here initial local_task is marked as steal and remote 
       task corresponds to a steal task 
       - push local_task into the own thread queue
       - push remote_task to the kproc' highpriority queue
       Note: the remote task is not visible in the local queue, because
       it was allocated into the data stack. 
    */
    kaapi_thread_pushtask(thread);
  }
#if 0 // BUGGY part: cannot be used because owner of the task does not execute them with queue_pop
  else 
  {
    /* non strict flag: push in both queues, the local queue of the running thread
       and the remote queue.
    */
    kaapi_sched_lock( &queue->lock );
    kaapi_queue_push( queue, local_task);
    kaapi_sched_unlock( &queue->lock );

    /* push the local task 
    */
    kaapi_thread_pushtask(thread);
  }
#endif

  return 0;
}


int kaapi_thread_distribute_task (
  kaapi_thread_t* thread,
  kaapi_processor_id_t kid
)
{
  //printf("[push::distribute_task] kid:%i\n", kid);
  return kaapi_thread_remotepush_task( thread, KAAPI_HWS_LEVELID_L0, kid, KAAPI_TASK_S_STRICT);
}

